# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# @ReservedAssignment

ctypedef unsigned char BYTE
ctypedef unsigned char CARD8
ctypedef unsigned short CARD16
ctypedef unsigned long CARD32
ctypedef int BOOL
ctypedef int Bool
ctypedef int Status
ctypedef CARD32 Atom
ctypedef CARD32 XID
ctypedef char* XPointer
ctypedef XID Drawable
ctypedef XID Window
ctypedef XID Pixmap
ctypedef XID Cursor
ctypedef CARD32 Colormap
ctypedef CARD32 VisualID
ctypedef CARD32 Time
ctypedef XID KeySym


cdef extern from "X11/X.h":
    unsigned long NoSymbol
    unsigned long AnyPropertyType
    unsigned int PropModeReplace
    unsigned int PropertyNotify
    unsigned int Expose


cdef extern from "X11/Xutil.h":
    ctypedef struct aspect:
        int x,y
    ctypedef struct XSizeHints:
        long flags                  #marks which fields in this structure are defined
        int x, y                    #Obsolete
        int width, height           #Obsolete
        int min_width, min_height
        int max_width, max_height
        int width_inc, height_inc
        aspect min_aspect, max_aspect
        int base_width, base_height
        int win_gravity
        #this structure may be extended in the future

    ctypedef struct XWMHints:
        long flags                  #marks which fields in this structure are defined
        Bool input                  #does this application rely on the window manager to get keyboard input?
        int initial_state
        Pixmap icon_pixmap          #pixmap to be used as icon
        Window icon_window          #window to be used as icon
        int icon_x, icon_y          #initial position of icon
        Pixmap icon_mask            #pixmap to be used as mask for icon_pixmap
        XID window_group            #id of related window group


cdef extern from "X11/Xlib.h":
    int CurrentTime
    int MappingBusy
    int GrabModeAsync
    int AnyKey
    int AnyModifier

    int VisualIDMask
    #query colors flags:
    int DoRed
    int DoGreen
    int DoBlue

    unsigned long AllPlanes
    int XYPixmap
    int ZPixmap
    int MSBFirst
    int LSBFirst

    int BadRequest
    int Success

    int XIAnyPropertyType

    ctypedef struct Display:
        pass

    Atom XInternAtom(Display * display, char * atom_name, Bool only_if_exists)
    Status XInternAtoms(Display *display, char **names, int count, Bool only_if_exists, Atom *atoms_return)
    char *XGetAtomName(Display *display, Atom atom)

    int XFree(void * data)
    int XKillClient(Display *, XID)

    void XGetErrorText(Display * display, int code, char * buffer_return, int length)

    int XUngrabKeyboard(Display * display, Time t)
    int XGrabPointer(Display *display, Window grab_window, Bool owner_events,
                     unsigned int event_mask, int pointer_mode, int keyboard_mode,
                     Window confine_to, Cursor cursor, Time time)
    int XUngrabPointer(Display *display, Time time)

    int *XSynchronize(Display *display, Bool onoff)

    Display *XOpenDisplay(char *display_name)
    int XCloseDisplay(Display *display)

    ctypedef struct XRectangle:
        short x, y
        unsigned short width, height

    ctypedef struct XClassHint:
        char *res_name
        char *res_class

    ctypedef struct XImage:
        int width
        int height
        int xoffset             # number of pixels offset in X direction
        int format              # XYBitmap, XYPixmap, ZPixmap
        char *data              # pointer to image data
        int byte_order          # data byte order, LSBFirst, MSBFirst
        int bitmap_unit         # quant. of scanline 8, 16, 32
        int bitmap_bit_order    # LSBFirst, MSBFirst
        int bitmap_pad          # 8, 16, 32 either XY or ZPixmap
        int depth               # depth of image
        int bytes_per_line      # accelerator to next scanline
        int bits_per_pixel      # bits per pixel (ZPixmap)
        unsigned long red_mask  # bits in z arrangement
        unsigned long green_mask
        unsigned long blue_mask
        XPointer *obdata
        void *funcs

    ctypedef struct Visual:
        void    *ext_data       #XExtData *ext_data;     /* hook for extension to hang data */
        VisualID visualid
        int c_class
        unsigned long red_mask
        unsigned long green_mask
        unsigned long blue_mask
        int bits_per_rgb
        int map_entries

    ctypedef struct XVisualInfo:
        Visual *visual
        VisualID visualid
        int screen
        unsigned int depth
        int c_class "class"
        unsigned long red_mask
        unsigned long green_mask
        unsigned long blue_mask
        int colormap_size
        int bits_per_rgb

    ctypedef struct XColor:
        unsigned long pixel                 # pixel value
        unsigned short red, green, blue     # rgb values
        char flags                          # DoRed, DoGreen, DoBlue

    ctypedef struct XSetWindowAttributes:
        Pixmap background_pixmap            # background, None, or ParentRelative
        unsigned long background_pixel      # background pixel
        Pixmap border_pixmap                # border of the window or CopyFromParent
        unsigned long border_pixel          # border pixel value
        int bit_gravity                     # one of bit gravity values
        int win_gravity                     # one of the window gravity values
        int backing_store                   # NotUseful, WhenMapped, Always
        unsigned long backing_planes        # planes to be preserved if possible
        unsigned long backing_pixel         # value to use in restoring planes
        Bool save_under                     # should bits under be saved? (popups)
        long event_mask                     # set of events that should be saved
        long do_not_propagate_mask          # set of events that should not propagate
        Bool override_redirect              # boolean value for override_redirect
        Colormap colormap                   # color map to be associated with window
        Cursor cursor

    ctypedef struct XWindowChanges:
        int x, y, width, height, border_width
        Window sibling
        int stack_mode

    ctypedef struct XWindowAttributes:
        int x, y, width, height, border_width
        int depth
        Visual *visual
        int _class "class"
        int bit_gravity, win_gravity, backing_store
        unsigned long backing_planes, backing_pixel
        Bool save_under
        Colormap colormap
        Bool map_installed
        int map_state
        long all_event_masks
        long your_event_mask
        long do_not_propagate_mask
        Bool override_redirect
        #Screen *screen

    ctypedef struct XGenericEventCookie:
        int            type     # of event. Always GenericEvent
        unsigned long  serial
        Bool           send_event
        Display        *display
        int            extension    #major opcode of extension that caused the event
        int            evtype       #actual event type
        unsigned int   cookie
        void           *data

    ctypedef unsigned char KeyCode
    ctypedef struct XModifierKeymap:
        int max_keypermod
        KeyCode * modifiermap # an array with 8*max_keypermod elements


    Bool XQueryExtension(Display * display, char *name,
                         int *major_opcode_return, int *first_event_return, int *first_error_return)

    Bool XGetEventData(Display *display, XGenericEventCookie *cookie)
    void XFreeEventData(Display *display, XGenericEventCookie *cookie)

    Window XDefaultRootWindow(Display * display)

    Bool XQueryPointer(Display *display, Window w, Window *root_return, Window *child_return, int *root_x_return, int *root_y_return,
                       int *win_x_return, int *win_y_return, unsigned int *mask_return)
    int XFlush(Display *dpy)
    int XSync(Display *dpy, Bool discard)

    int XConnectionNumber(Display *display)

    int XPending(Display *display)

    # Icon sizes
    ctypedef struct XIconSize:
        int min_width
        int min_height
        int max_width
        int max_height
        int width_inc;
        int height_inc;

    XIconSize *XAllocIconSize()
    int XSetIconSizes(Display *display, Window w, XIconSize* size_list, int count)

    # Keyboard bindings
    XModifierKeymap* XGetModifierMapping(Display* display)
    int XFreeModifiermap(XModifierKeymap* modifiermap)
    int XDisplayKeycodes(Display* display, int* min_keycodes, int* max_keycodes)
    KeySym XStringToKeysym(char* string)
    KeySym* XGetKeyboardMapping(Display* display, KeyCode first_keycode, int keycode_count, int* keysyms_per_keycode_return)
    int XChangeKeyboardMapping(Display* display, int first_keycode, int keysyms_per_keycode, KeySym* keysyms, int num_codes)
    XModifierKeymap* XInsertModifiermapEntry(XModifierKeymap* modifiermap, KeyCode keycode_entry, int modifier)
    char* XKeysymToString(KeySym keysym)

    int XSetModifierMapping(Display* display, XModifierKeymap* modifiermap)

    int XGrabKey(Display * display, int keycode, unsigned int modifiers,
                 Window grab_window, Bool owner_events,
                 int pointer_mode, int keyboard_mode)
    int XUngrabKey(Display * display, int keycode, unsigned int modifiers,
                   Window grab_window)
    int XQueryKeymap(Display * display, char [32] keys_return)

    #Threading:
    Status XInitThreads()

    # error handling:
    ctypedef int (*X11IOERRORHANDLER)(Display *) except 0
    int XSetIOErrorHandler(X11IOERRORHANDLER  handler)
    ctypedef struct XErrorEvent:
        int type
        Display *display
        unsigned long serial
        unsigned char error_code
        unsigned char request_code
        unsigned char minor_code
        XID resourceid
    ctypedef int (*X11ERRORHANDLER)(Display *, XErrorEvent *event) except 0
    int XSetErrorHandler(X11ERRORHANDLER handler)

    # events:
    Status XSendEvent(Display *, Window target, Bool propagate,
                      unsigned long event_mask, XEvent * event)
    void XNextEvent(Display *display, XEvent *event_return) nogil
    int XSelectInput(Display * display, Window w, unsigned long event_mask)

    # properties:
    int XChangeProperty(Display *, Window w, Atom prop,
         Atom ptype, int fmt, int mode, unsigned char * data, int nelements)
    int XGetWindowProperty(Display * display, Window w, Atom prop,
         long offset, long length, Bool delete,
         Atom req_type, Atom * actual_type,
         int * actual_format,
         unsigned long * nitems, unsigned long * bytes_after,
         unsigned char ** prop)
    int XDeleteProperty(Display * display, Window w, Atom prop)

    int XAddToSaveSet(Display *, Window w)
    int XRemoveFromSaveSet(Display *, Window w)

    Visual *XDefaultVisual(Display *display, int screen_number)
    XVisualInfo *XGetVisualInfo(Display *display, long vinfo_mask, XVisualInfo *vinfo_template, int *nitems_return)

    # windows:
    Status XGetWindowAttributes(Display * display, Window w,
                                XWindowAttributes * attributes)

    Window XCreateWindow(Display *display, Window parent,
                         int x, int y, unsigned int width, unsigned int height, unsigned int border_width, int depth,
                         unsigned int _class, Visual *visual,
                         unsigned long valuemask, XSetWindowAttributes *attributes)
    int XDestroyWindow(Display *display, Window w)

    int XConfigureWindow(Display * display, Window w,
         unsigned int value_mask, XWindowChanges * changes)
    Status XReconfigureWMWindow(Display * display, Window w, int screen_number,
                                unsigned int value_mask, XWindowChanges *values)
    int XMoveResizeWindow(Display * display, Window w, int x, int y, int width, int height)
    int XMoveWindow(Display * display, Window w, int x, int y)
    int XResizeWindow(Display * display, Window w, int width, int height)

    Bool XTranslateCoordinates(Display * display,
                               Window src_w, Window dest_w,
                               int src_x, int src_y,
                               int * dest_x, int * dest_y,
                               Window * child)

    Status XQueryTree(Display * display, Window w,
                      Window * root, Window * parent,
                      Window ** children, unsigned int * nchildren)

    # focus:
    int XSetInputFocus(Display * display, Window focus,
                                          int revert_to, Time ctime)
    int XGetInputFocus(Display * display, Window * focus,
                                          int * revert_to)

    # XUnmapWindow
    int XUnmapWindow(Display *, Window)
    unsigned long NextRequest(Display *)

    int XIconifyWindow(Display *, Window, int screen_number)

    # XMapWindow
    int XMapWindow(Display *, Window)
    int XMapRaised(Display *, Window)
    Status XWithdrawWindow(Display *, Window, int screen_number)
    void XReparentWindow(Display *, Window w, Window parent, int x, int y)
    void XRaiseWindow(Display *display, Window w)

    XClassHint *XAllocClassHint()
    Status XGetClassHint(Display *display, Window w, XClassHint *class_hints_return)
    void XSetClassHint(Display *display, Window w, XClassHint *class_hints)

    Status XGetGeometry(Display *display, Drawable d, Window *root_return,
                        int *x_return, int *y_return, unsigned int  *width_return, unsigned int *height_return,
                        unsigned int *border_width_return, unsigned int *depth_return)

    XSizeHints *XAllocSizeHints()
    #Status XGetWMSizeHints(Display *display, Window w, XSizeHints *hints_return, long *supplied_return, Atom property)
    void XSetWMNormalHints(Display *display, Window w, XSizeHints *hints)
    Status XGetWMNormalHints(Display *display, Window w, XSizeHints *hints_return, long *supplied_return)
    XWMHints *XGetWMHints(Display *display, Window w)

    Status XGetWMProtocols(Display *display, Window w, Atom **protocols_return, int *count_return)

    # selection:
    Window XGetSelectionOwner(Display * display, Atom selection)
    int XSetSelectionOwner(Display * display, Atom selection, Window owner, Time ctime)
    int XConvertSelection(Display * display, Atom selection, Atom target, Atom prop,
                          Window requestor, Time time)

    #ctypedef Bool (*XIf_predicate) (Display *display, XEvent *event, XPointer arg) nogil
    int XIfEvent(Display *display, XEvent *event_return, void *predicate, XPointer arg)

    # There are way more event types than this; add them as needed.
    ctypedef struct XAnyEvent:
        int type
        unsigned long serial
        Bool send_event
        Display * display
        Window window
    # Needed to broadcast that we are a window manager, among other things:
    union payload_for_XClientMessageEvent:
        char b[20]
        short s[10]
        unsigned long l[5]
    ctypedef struct XClientMessageEvent:
        Atom message_type
        int format
        payload_for_XClientMessageEvent data
    # SubstructureRedirect-related events:
    ctypedef struct XMapRequestEvent:
        Window parent  # Same as xany.window, confusingly.
        Window window
    ctypedef struct XConfigureRequestEvent:
        Window parent  # Same as xany.window, confusingly.
        Window window
        int x, y, width, height, border_width
        Window above
        int detail
        unsigned long value_mask
    ctypedef struct XSelectionRequestEvent:
        Window owner
        Window requestor
        Atom selection
        Atom target
        Atom property
        Time time
    ctypedef struct XSelectionClearEvent:
        Window window
        Atom selection
        Time time
    ctypedef struct XSelectionEvent:
        int subtype
        Window requestor
        Atom selection
        Atom target
        Atom property
        Time time
    ctypedef struct XResizeRequestEvent:
        Window window
        int width, height
    ctypedef struct XReparentEvent:
        Window window
        Window parent
        int x, y
    ctypedef struct XCirculateRequestEvent:
        Window parent  # Same as xany.window, confusingly.
        Window window
        int place
    # For pointer grabs:
    ctypedef struct XCrossingEvent:
        unsigned long serial
        Bool send_event
        Window window
        Window root
        Window subwindow
        int mode                # NotifyNormal, NotifyGrab, NotifyUngrab
        int detail              # NotifyAncestor, NotifyVirtual, NotifyInferior, NotifyNonlinear,NotifyNonlinearVirtual
        Bool focus
        unsigned int state
    # Focus handling
    ctypedef struct XFocusChangeEvent:
        Window window
        int mode                #NotifyNormal, NotifyGrab, NotifyUngrab
        int detail              #NotifyAncestor, NotifyVirtual, NotifyInferior,
                                #NotifyNonlinear,NotifyNonlinearVirtual, NotifyPointer,
                                #NotifyPointerRoot, NotifyDetailNone
    ctypedef struct XMotionEvent:
        Window window           #event window reported relative to
        Window root             #root window that the event occurred on
        Window subwindow        #child window
        Time time               #milliseconds
        int x, y                #pointer x, y coordinates in event window
        int x_root, y_root      #coordinates relative to root
        unsigned int state      #key or button mask
        char is_hint            #detail
        Bool same_screen        #same screen
    # We have to generate synthetic ConfigureNotify's:
    ctypedef struct XConfigureEvent:
        Window event    # Same as xany.window, confusingly.
                        # The selected-on window.
        Window window   # The effected window.
        int x, y, width, height, border_width
        Window above
        Bool override_redirect
    ctypedef struct XCreateWindowEvent:
        Window window
        int width
        int height
    ctypedef struct XButtonEvent:
        Window root
        Window subwindow
        Time time
        int x, y                # pointer x, y coordinates in event window
        int x_root, y_root      # coordinates relative to root */
        unsigned int state      # key or button mask
        unsigned int button
        Bool same_screen
    ctypedef struct XMapEvent:
        Window window
        Bool override_redirect
    ctypedef struct XUnmapEvent:
        Window window
        Bool from_configure
    ctypedef struct XDestroyWindowEvent:
        Window window
    ctypedef struct XPropertyEvent:
        Window window
        Atom atom
        Time time
    ctypedef struct XKeyEvent:
        unsigned int state
        unsigned int keycode

    ctypedef struct XExposeEvent:
        Window window
        int x, y
        int width, height
        int count

    ctypedef union XEvent:
        int type
        XAnyEvent xany
        XKeyEvent xkey
        XButtonEvent xbutton
        XMapRequestEvent xmaprequest
        XConfigureRequestEvent xconfigurerequest
        XSelectionRequestEvent xselectionrequest
        XSelectionClearEvent xselectionclear
        XResizeRequestEvent xresizerequest
        XCirculateRequestEvent xcirculaterequest
        XConfigureEvent xconfigure
        XCrossingEvent xcrossing
        XFocusChangeEvent xfocus
        XMotionEvent xmotion
        XClientMessageEvent xclient
        XSelectionEvent xselection
        XMapEvent xmap
        XCreateWindowEvent xcreatewindow
        XUnmapEvent xunmap
        XReparentEvent xreparent
        XDestroyWindowEvent xdestroywindow
        XPropertyEvent xproperty
        XGenericEventCookie xcookie
        XExposeEvent xexpose



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import sys
import signal
from typing import Any
from subprocess import Popen, PIPE

from xpra.os_util import POSIX, LINUX, OSX, WIN32
from xpra.util.env import _saved_env, envbool
from xpra.util.io import load_binary_file, get_util_logger


SIGNAMES: dict[int, str] = {}
for signame in (sig for sig in dir(signal) if sig.startswith("SIG") and not sig.startswith("SIG_")):
    try:
        SIGNAMES[int(getattr(signal, signame))] = signame
    except ValueError:
        pass


def set_proc_title(title: str) -> None:
    try:
        import setproctitle  # pylint: disable=import-outside-toplevel
        setproctitle.setproctitle(title)  # @UndefinedVariable pylint: disable=c-extension-no-member
    except ImportError as e:
        get_util_logger().debug("setproctitle is not installed: %s", e)


def no_idle(fn, *args, **kwargs) -> None:
    fn(*args, **kwargs)


def register_SIGUSR_signals(idle_add=no_idle) -> None:
    if os.name != "posix":
        return
    from xpra.util.pysystem import dump_gc_frames
    from xpra.util.pysystem import dump_all_frames

    def sigusr1(*_args) -> None:
        log = get_util_logger().info
        log("SIGUSR1")
        idle_add(dump_all_frames, log)

    def sigusr2(*_args) -> None:
        log = get_util_logger().info
        log("SIGUSR2")
        idle_add(dump_gc_frames, log)

    signal.signal(signal.SIGUSR1, sigusr1)
    signal.signal(signal.SIGUSR2, sigusr2)


def is_VirtualBox() -> bool:
    if os.name != "nt":
        # detection not supported yet
        return False
    # try to detect VirtualBox:
    # based on the code found here:
    # http://spth.virii.lu/eof2/articles/WarGame/vboxdetect.html
    try:
        from ctypes import cdll
        if cdll.LoadLibrary("VBoxHook.dll"):
            # "VirtualBox is present (VBoxHook.dll)
            return True
    except (ImportError, OSError):
        pass
    try:
        with open("\\\\.\\VBoxMiniRdrDN", "r"):
            # "VirtualBox is present (VBoxMiniRdrDN)"
            return True
    except Exception as e:
        import errno
        if e.args[0] == errno.EACCES:
            # "VirtualBox is present (VBoxMiniRdrDN)"
            return True
    return False


def is_Wayland() -> bool:
    return _is_Wayland(_saved_env)


def _is_Wayland(env: dict) -> bool:
    backend = env.get("GDK_BACKEND", "")
    if backend == "wayland":
        return True
    return backend != "x11" and (
        bool(env.get("WAYLAND_DISPLAY")) or env.get("XDG_SESSION_TYPE") == "wayland"
    )


def is_distribution_variant(variant="Debian") -> bool:
    if not POSIX:
        return False
    try:
        v = load_os_release_file()
        return any(line.find(variant) >= 0 for line in v.splitlines() if line.startswith("NAME="))
    except Exception:
        pass
    try:
        d = get_linux_distribution()[0]
        if d == variant:
            return True
        if variant == "RedHat" and d.startswith(variant):
            return True
    except Exception:
        pass
    return False


def get_distribution_version_id() -> str:
    if not POSIX:
        return ""
    try:
        v = load_os_release_file()
        for bline in v.splitlines():
            line = bline.decode()
            if line.startswith("VERSION_ID="):
                return line.split("=", 1)[1].strip('"')
    except Exception:
        pass
    return ""


os_release_file_data: str | None = None


def load_os_release_file() -> str:
    global os_release_file_data
    if os_release_file_data is None:
        try:
            os_release_file_data = load_binary_file("/etc/os-release").decode() or ""
        except (OSError, UnicodeDecodeError):  # pragma: no cover
            os_release_file_data = ""
    return os_release_file_data


def is_Ubuntu() -> bool:
    return is_distribution_variant("Ubuntu")


def is_LinuxMint() -> bool:
    return is_distribution_variant("Linux Mint")


def is_Debian() -> bool:
    return is_distribution_variant("Debian")


def is_DEB() -> bool:
    return is_Debian() or is_Ubuntu() or is_LinuxMint()


def is_RPM() -> bool:
    return any(is_distribution_variant(x) for x in (
        "Fedora", "RedHat", "AlmaLinux", "Rocky Linux", "CentOS", "openSUSE", "Oracle Linux",
    ))


_linux_distribution = ("", "", "")


def get_linux_distribution() -> tuple[str, str, str]:
    global _linux_distribution
    if LINUX and _linux_distribution == ("", "", ""):
        # linux_distribution is deprecated in Python 3.5,
        # and it causes warnings,
        # so we use our own code first:
        cmd = ["lsb_release", "-a"]
        try:
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
            out = p.communicate()[0]
            assert p.returncode == 0 and out
        except Exception:
            try:
                import platform
                _linux_distribution = platform.linux_distribution()  # pylint: disable=deprecated-method, no-member
            except Exception:
                _linux_distribution = ("unknown", "unknown", "unknown")
        else:
            d = {}
            for line in out.splitlines():
                parts = line.rstrip("\n\r").split(":", 1)
                if len(parts) == 2:
                    d[parts[0].lower().replace(" ", "_")] = parts[1].strip()
            _linux_distribution = (
                d.get("distributor_id", "unknown"),
                d.get("release", "unknown"),
                d.get("codename", "unknown"),
            )
    return _linux_distribution


def is_unity() -> bool:
    d = os.environ.get("XDG_CURRENT_DESKTOP", "").lower()
    return d.find("unity") >= 0 or d.find("ubuntu") >= 0


def is_gnome() -> bool:
    if os.environ.get("XDG_SESSION_DESKTOP", "").split("-", 1)[0] in ("i3", "ubuntu",):
        # "i3-gnome" is not really gnome... ie: the systray does work!
        return False
    return os.environ.get("XDG_CURRENT_DESKTOP", "").lower().find("gnome") >= 0


def is_kde() -> bool:
    return os.environ.get("XDG_CURRENT_DESKTOP", "").lower().find("kde") >= 0


def get_loaded_kernel_modules(*modlist):
    loaded = []
    if LINUX and os.path.exists("/sys/module"):
        for mod in modlist:
            if os.path.exists(f"/sys/module/{mod}"):  # pragma: no cover
                loaded.append(mod)
    return loaded


def is_WSL() -> bool:
    if not POSIX:
        return False
    for f in ("/proc/sys/kernel/osrelease", "/proc/version"):
        r = load_binary_file(f)
        if r:
            return r.find(b"Microsoft") >= 0
    return False


def get_generic_os_name() -> str:
    return do_get_generic_os_name().lower()


def do_get_generic_os_name() -> str:
    for k, v in {
        "linux": "Linux",
        "darwin": "MacOS",
        "win": "MS Windows",
        "freebsd": "FreeBSD",
    }.items():
        if sys.platform.startswith(k):
            return v
    return sys.platform  # pragma: no cover


def is_X11() -> bool:
    if OSX or WIN32:
        return False
    if envbool("XPRA_NOX11", False):
        return False
    try:
        from xpra import x11
        assert x11
    except ImportError:
        # x11 is not installed, so assume it isn't used
        return False
    try:
        from xpra.x11.gtk.bindings import is_X11_Display
        return is_X11_Display()
    except ImportError:
        get_util_logger().debug("failed to load x11 bindings", exc_info=True)
        return True


def nn(x) -> str:
    if x is None:
        return ""
    return str(x)


def get_env_info() -> dict[str, str]:
    filtered_env = os.environ.copy()
    if filtered_env.get('XPRA_PASSWORD'):
        filtered_env['XPRA_PASSWORD'] = "*****"
    if filtered_env.get('XPRA_ENCRYPTION_KEY'):
        filtered_env['XPRA_ENCRYPTION_KEY'] = "*****"
    return filtered_env


def get_sysconfig_info() -> dict[str, Any]:
    import sysconfig
    sysinfo: dict[str, Any] = {}
    log = get_util_logger()
    for attr in (
            "platform",
            "python-version",
            "config-vars",
            "paths",
    ):
        fn = "get_" + attr.replace("-", "_")
        getter = getattr(sysconfig, fn, None)
        if getter:
            try:
                sysinfo[attr] = getter()  # pylint: disable=not-callable
            except ModuleNotFoundError:
                log("sysconfig.%s", fn, exc_info=True)
                if attr == "config-vars" and WIN32:
                    continue
                log.warn("Warning: failed to collect %s sysconfig information", attr)
            except Exception:
                log.error("Error calling sysconfig.%s", fn, exc_info=True)
    return sysinfo


def platform_release(release):
    log = get_util_logger()
    if WIN32:
        try:
            import wmi
        except ImportError:
            log(f"platform_release({release}) no wmi", exc_info=True)
            if release.endswith("Server"):
                return release.replace("Server", "-Server")  # ie: "2025Server" -> "2025-Server"
        else:
            try:
                computer = wmi.WMI()
                os_info = computer.Win32_OperatingSystem()[0]
                return os_info.Name.split('|')[0]
            except Exception as e:
                log.debug("wmi query", exc_info=True)
                log.warn("Warning: failed to query OS using wmi")
                log.warn(f" {e}")
    if OSX:
        systemversion_plist = "/System/Library/CoreServices/SystemVersion.plist"
        try:
            import plistlib
            with open(systemversion_plist, "rb") as f:
                pl = plistlib.load(f)
            return pl['ProductUserVisibleVersion']
        except Exception as e:
            log.debug("platform_release(%s)", release, exc_info=True)
            log.warn("Warning: failed to get release information")
            log.warn(f" from {systemversion_plist}:")
            log.warn(f" {e}")
    return release


def platform_name(sys_platform=sys.platform, release="") -> str:
    if not sys_platform:
        return "unknown"
    platforms = {
        "win32": "Microsoft Windows",
        "cygwin": "Windows/Cygwin",
        "linux.*": "Linux",
        "darwin": "Mac OS X",
        "freebsd.*": "FreeBSD",
        "os2": "OS/2",
    }

    def rel(v):
        if isinstance(release, (tuple, list)):
            values = [v] + list(release)
        else:
            values = []
            if not release.startswith(v):
                values.append(v)
            values.append(release)
        return " ".join(str(x) for x in values if x and x not in ("", "unknown", "n/a"))

    for k, v in platforms.items():
        regexp = re.compile(k)
        if regexp.match(sys_platform):
            return rel(v)
    return rel(sys_platform)


def is_systemd_pid1() -> bool:
    if not POSIX:
        return False
    d = load_binary_file("/proc/1/cmdline")
    return d.find(b"/systemd") >= 0


def get_processor_name() -> str:
    import platform
    sysname = platform.system()
    if sysname == "Windows":
        return platform.processor()
    if sysname == "Darwin":
        from xpra.util.env import OSEnvContext
        from subprocess import check_output
        with OSEnvContext():
            os.environ['PATH'] = os.environ['PATH'] + os.pathsep + '/usr/sbin'
            command = ["sysctl", "-n", "machdep.cpu.brand_string"]
            return check_output(command, text=True).strip()
    if sysname == "Linux":
        with open("/proc/cpuinfo", encoding="latin1") as f:
            data = f.read()
        import re
        for line in data.split("\n"):
            if "model name" in line:
                return re.sub(".*model name.*:", "", line, count=1).strip()
    return platform.processor()



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable

from xpra.util.objects import typedict
from xpra.os_util import gi_import
from xpra.net.protocol.factory import get_server_protocol_class
from xpra.server.proxy.instance_base import ProxyInstance
from xpra.log import Logger

log = Logger("proxy")

GLib = gi_import("GLib")


class ProxyInstanceThread(ProxyInstance):

    def __init__(self, session_options: dict[str, str], pings: int,
                 client_proto, server_conn,
                 disp_desc: dict[str, Any],
                 cipher: str, cipher_mode: str, encryption_key: bytes, caps: typedict):
        super().__init__(session_options, pings,
                         disp_desc, cipher, cipher_mode, encryption_key, caps)
        self.client_protocol = client_proto
        self.server_conn = server_conn

    def __repr__(self):
        return "threaded proxy instance"

    def idle_add(self, fn: Callable, *args, **kwargs) -> int:
        return GLib.idle_add(fn, *args, **kwargs)

    def timeout_add(self, timeout, fn: Callable, *args, **kwargs) -> int:
        return GLib.timeout_add(timeout, fn, *args, **kwargs)

    def source_remove(self, tid: int) -> None:
        GLib.source_remove(tid)

    def run(self) -> None:
        log("ProxyInstanceThread.run()")
        server_protocol_class = get_server_protocol_class(self.server_conn.socktype)
        self.server_protocol = server_protocol_class(self.server_conn,
                                                     self.process_server_packet, self.get_server_packet)
        self.log_start()
        super().run()

    def start_network_threads(self) -> None:
        log("start_network_threads()")
        self.server_protocol.start()
        self.client_protocol._process_packet_cb = self.process_client_packet
        self.client_protocol.set_packet_source(self.get_client_packet)
        # no need to start the client protocol,
        # it was started when we processed authentication in the proxy server
        # self.client_protocol.start()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from time import monotonic

from xpra.util.str_fn import hexstr
from xpra.codecs.argb.argb import r210_to_rgba, r210_to_rgbx, argb_to_rgba, bgra_to_rgba     # pylint: disable=no-name-in-module


def measure_fn(fn, data, *args):
    N = 10
    start = monotonic()
    for _ in range(N):
        fn(data, *args)
    end = monotonic()
    mps = len(data)*N//4/(end-start)//1024//1024
    print(f"{fn}: {mps} MPixels/s")


def cmp(inbytes, outbytes, fn, *args):
    datain = bytes(bytearray(inbytes))
    dataout = bytes(bytearray(outbytes))
    r = bytes(bytearray(fn(datain, *args)))
    assert dataout==r, f"expected {hexstr(dataout)} but got {hexstr(r)}"


class ARGBTest(unittest.TestCase):

    def test_r210_to_rgba(self):
        cmp((0xff, 0xfe, 0x7f, 0x7e),
            (0xf9, 0xff, 0xbf, 0x55),
            r210_to_rgba, 1, 1, 4, 4,
            )
        cmp((0x17, 0x0f, 0x31, 0x8f),
            (0x3c, 0x10, 0xc5, 0xaa),
            r210_to_rgba, 1, 1, 4, 4,
            )
        w = 1920
        h = 1080
        data = bytes(bytearray(w*h*4))
        measure_fn(r210_to_rgba, data, w, h, w*4, w*4)

    def test_r210data_to_rgbx(self):
        cmp((0xff, 0xfe, 0x7f, 0x7e),
            (0xf9, 0xff, 0xbf, 0xff),
            r210_to_rgbx, 1, 1, 4, 4,
            )
        cmp((0x17, 0x0f, 0x31, 0x8f),
            (0x3c, 0x10, 0xc5, 0xff),
            r210_to_rgbx, 1, 1, 4, 4,
            )
        w = 1920
        h = 1080
        data = bytes(bytearray(w*h*4))
        measure_fn(r210_to_rgbx, data, w, h, w*4, w*4)

    def test_argb_to_rgba(self):
        cmp((0xff, 0xfe, 0x7f, 0x7e),
            (0xfe, 0x7f, 0x7e, 0xff),
            argb_to_rgba,
            )
        cmp((0x17, 0x0f, 0x31, 0x8f),
            (0x0f, 0x31, 0x8f, 0x17),
            argb_to_rgba,
            )
        w = 1920
        h = 1080
        data = bytes(bytearray(w*h*4))
        measure_fn(argb_to_rgba, data)

    def test_bgra_to_rgba(self):
        cmp((0xff, 0xfe, 0x7f, 0x7e),
            (0x7f, 0xfe, 0xff, 0x7e),
            bgra_to_rgba,
            )
        cmp((0x17, 0x0f, 0x31, 0x8f),
            (0x31, 0x0f, 0x17, 0x8f),
            bgra_to_rgba,
            )
        w = 1920
        h = 1080
        data = bytes(bytearray(w*h*4))
        measure_fn(bgra_to_rgba, data)


def main():
    unittest.main()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.x11.bindings.xlib cimport (
    Display, XID, Atom, Bool, Status, Window, CARD32,
    Success,
)
from xpra.x11.bindings.core cimport X11CoreBindingsInstance

from xpra.log import Logger
log = Logger("x11", "bindings")

ctypedef int pid_t

###################################
# XRes
###################################
cdef extern from "X11/extensions/XRes.h":
    ctypedef enum XResClientIdType:
        XRES_CLIENT_ID_XID
        XRES_CLIENT_ID_PID
        XRES_CLIENT_ID_NR

    ctypedef enum XResClientIdMask:
        XRES_CLIENT_ID_XID_MASK = 1 << XRES_CLIENT_ID_XID
        XRES_CLIENT_ID_PID_MASK = 1 << XRES_CLIENT_ID_PID

    ctypedef struct XResClientIdSpec:
        XID client
        unsigned int mask

    ctypedef struct XResResourceIdSpec:
        XID resource
        Atom type

    ctypedef struct XResClientIdValue:
        XResClientIdSpec spec
        long          length
        void         *value

    ctypedef struct XResClient:
        XID    resource_base
        XID    resource_mask

    ctypedef struct XResType:
        Atom    resource_type
        unsigned int    count

    Status XResQueryClientIds(Display *dpy, long num_specs, XResClientIdSpec *client_specs,
                              long *num_ids, XResClientIdValue **client_ids)
    void XResClientIdsDestroy(long num_ids, XResClientIdValue  *client_ids)
    XResClientIdType XResGetClientIdType(XResClientIdValue* value)
    pid_t XResGetClientPid(XResClientIdValue* value)

    Bool XResQueryExtension(Display *dpy, int *event_base_return, int *error_base_return)
    Status XResQueryVersion(Display *dpy, int *major_version_return, int *minor_version_return)


cdef unsigned int get_pid(Display *display, Window xid) noexcept:
    cdef XResClientIdSpec client_spec
    client_spec.client = xid
    client_spec.mask = XRES_CLIENT_ID_PID_MASK

    cdef long num_ids
    cdef XResClientIdValue *client_ids
    if XResQueryClientIds(display, 1, &client_spec, &num_ids, &client_ids):
        log.error("Error: failed to query pid for window %i", xid)
        return 0

    cdef int pid = 0
    for i in range(num_ids):
        if client_ids[i].spec.mask == XRES_CLIENT_ID_PID_MASK:
            pid = XResGetClientPid(&client_ids[i])
            if pid>=0:
                break
    if num_ids:
        XResClientIdsDestroy(num_ids, client_ids)
    return pid


cdef class ResBindingsInstance(X11CoreBindingsInstance):

    def __init__(self):
        pass

    def __repr__(self):
        return "XResBindings(%s)" % self.display_name

    def check_xres(self, min_version=(1, 2)) -> bool:
        cdef int event_base = 0, ignored = 0, cmajor = 0, cminor = 0
        cdef Bool r = XResQueryExtension(self.display, &event_base, &ignored)
        log("XResQueryExtension()=%i", r)
        if not r:
            return False
        log("found XRes extension")
        if not XResQueryVersion(self.display, &cmajor, &cminor):
            return False
        log("found XRes extension version %i.%i", cmajor, cminor)
        return (cmajor, cminor) >= min_version

    def get_pid(self, Window xid) -> long:
        self.context_check("get_pid")
        return get_pid(self.display, xid)


cdef ResBindingsInstance singleton = None


def ResBindings() -> ResBindingsInstance:
    global singleton
    if singleton is None:
        singleton = ResBindingsInstance()
    return singleton



#!/usr/bin/env python3

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GLib	#pylint: disable=wrong-import-position @UnresolvedImport


def make_window(OR=False):
	class State(object):
		pass
	state = State()
	state.counter = 0
	state.width = 400-OR*100
	state.height = 200-OR*50
	window = Gtk.Window()
	window.set_size_request(state.width, state.height)
	window.connect("delete_event", Gtk.main_quit)
	window.set_title("OR=%s" % OR)
	window.set_position(Gtk.WindowPosition.CENTER)
	vbox = Gtk.VBox()
	hbox = Gtk.HBox()
	w_e = Gtk.Entry()
	w_e.set_max_length(64)
	hbox.add(w_e)
	hbox.add(Gtk.Label(label="x"))
	h_e = Gtk.Entry()
	h_e.set_max_length(64)
	hbox.add(h_e)
	set_size_btn = Gtk.Button("set size")
	def set_size(*_args):
		w = int(w_e.get_text())
		h = int(h_e.get_text())
		print("resizing to %s x %s" % (w, h))
		window.resize(w, h)
	set_size_btn.connect('clicked', set_size)
	hbox.add(set_size_btn)
	vbox.add(hbox)

	btn = Gtk.Button(label="auto resize me")
	def resize(*_args):
		state.width = max(200, (state.width+20) % 600)
		state.height = max(200, (state.height+20) % 400)
		print("resizing to %s x %s" % (state.width, state.height))
		window.resize(state.width, state.height)
	btn.connect('clicked', resize)
	vbox.add(btn)

	btn = Gtk.Button(label="fast resize")
	def do_resize_fast():
		width = max(200, (state.width+1) % 600)
		height = max(200, (state.height+1) % 400)
		print("resizing to %s x %s" % (width, height))
		window.resize(width, height)
		state.counter -= 1
		return state.counter>0
	def resize_fast(*_args):
		state.counter = 200
		GLib.idle_add(do_resize_fast)
	btn.connect('clicked', resize_fast)
	vbox.add(btn)

	window.add(vbox)
	window.realize()
	window.get_window().set_override_redirect(OR)
	window.show_all()

def main():
	make_window(False)
	make_window(True)
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shutil
import unittest
import tempfile
from time import monotonic
from subprocess import Popen

from xpra.util.str_fn import repr_ellipsized
from xpra.util.env import envint
from xpra.os_util import OSX, POSIX
from xpra.util.io import load_binary_file, pollwait
from xpra.exit_codes import ExitCode, ExitValue
from xpra.platform.dotxpra import DISPLAY_PREFIX
from unit.test_util import get_free_tcp_port
from unit.server_test_util import ServerTestUtil, log, estr, log_gap

CONNECT_WAIT = envint("XPRA_TEST_CONNECT_WAIT", 60)
SUBPROCESS_WAIT = envint("XPRA_TEST_SUBPROCESS_WAIT", CONNECT_WAIT * 2)

NOVERIFY = "--ssl-server-verify-mode=none"
NOHOSTNAME = "--ssl-check-hostname=no"


class GenSSLCertContext:
    __slots__ = ("tmpdir", "keyfile", "outfile", "certfile")

    def __init__(self):
        self._clear()

    def _clear(self):
        self.tmpdir = self.keyfile = self.outfile = self.certfile = ""

    def __enter__(self):
        self.tmpdir = tempfile.mkdtemp(suffix='ssl-xpra')
        self.keyfile = os.path.join(self.tmpdir, "key.pem")
        self.outfile = os.path.join(self.tmpdir, "out.pem")
        openssl_command = [
            "openssl", "req", "-new", "-newkey", "rsa:4096", "-days", "2", "-nodes", "-x509",
            "-subj", "/C=US/ST=Denial/L=Springfield/O=Dis/CN=localhost",
            "-keyout", self.keyfile, "-out", self.outfile,
        ]
        proc = Popen(args=openssl_command)
        assert pollwait(proc, 20) == 0, "openssl certificate generation failed"
        # combine the two files:
        self.certfile = os.path.join(self.tmpdir, "cert.pem")
        with open(self.certfile, "wb") as cf:
            for fname in (self.keyfile, self.outfile):
                with open(fname, "rb") as f:
                    cf.write(f.read())
        cert_data = load_binary_file(self.certfile)
        log("generated cert data: %s", repr_ellipsized(cert_data))
        if not cert_data:
            # cannot run openssl? (happens from rpmbuild)
            raise RuntimeError("SSL test skipped, cannot run " + " ".join(openssl_command))
        return self

    def __exit__(self, *_args):
        for f in (self.keyfile, self.outfile, self.certfile):
            os.unlink(f)
        os.rmdir(self.tmpdir)
        self._clear()

    def __repr__(self):
        return "GenSSLCertContext"


class ServerSocketsTest(ServerTestUtil):

    @classmethod
    def setUpClass(cls):
        ServerTestUtil.setUpClass()
        cls.default_xpra_args += [
            "--start-new-commands=no",
            "--video-encoders=none",
            "--csc-modules=none",
            "--video-decoders=none",
            "--encodings=rgb",
            "--mdns=no",
            "--webcam=no",
        ]

    def get_run_env(self):
        env = super().get_run_env()
        env["XPRA_CONNECT_TIMEOUT"] = str(CONNECT_WAIT)
        return env

    def start_server(self, *args):
        server_proc = self.run_xpra(["start", "--no-daemon"] + list(args))
        if pollwait(server_proc, 10) is not None:
            r = server_proc.poll()
            raise Exception(f"server failed to start with args={args}, returned {estr(r)}")
        return server_proc

    def _test_connect(self, server_args=(), client_args=(), password=None, uri_prefix=DISPLAY_PREFIX, exit_code=0):
        display_no = self.find_free_display_no()
        display = f":{display_no}"
        log(f"starting test server on {display}")
        server = self.start_server(display, "--printing=no", *server_args)
        # we should always be able to get the version:
        uri = uri_prefix + str(display_no)
        start = monotonic()
        while True:
            args = ["version", uri] + list(client_args or ())
            client = self.run_xpra(args)
            r = pollwait(client, CONNECT_WAIT)
            if r == 0:
                break
            if r is None:
                client.terminate()
            if monotonic() - start > SUBPROCESS_WAIT:
                if exit_code == ExitCode.CONNECTION_FAILED:
                    return
                err_msg = f"version client failed to connect using {args}, returned {estr(r)}"
                log.error(err_msg)
                log.error(f" server was started on {display=} with {server_args}")
                raise Exception(err_msg)
        # try to connect
        cmd = ["connect-test", uri] + [x.replace("$DISPLAY_NO", str(display_no)) for x in client_args]
        f = None
        if password:
            f = self._temp_file(password)
            cmd += [f"--password-file={f.name}"]
            cmd += [f"--challenge-handlers=file:filename={f.name}"]
        client = self.run_xpra(cmd)
        r = pollwait(client, SUBPROCESS_WAIT)
        if f:
            f.close()
        if r is None:
            client.terminate()
        server.terminate()
        if r != exit_code:
            log.error("Exit code mismatch")
            log.error(" expected %s (%s)", estr(exit_code), exit_code)
            log.error(" got %s (%s)", estr(r), r)
            log.error(" server args=%s", server_args)
            log.error(" client args=%s", client_args)
            self.verify_exitcode(expected=exit_code, actual=r)
        pollwait(server, 10)

    def test_default_socket(self):
        self._test_connect(["--bind=auto,auth=allow"], [], b"hello", DISPLAY_PREFIX, ExitCode.OK)

    def test_tcp_socket(self):
        port = get_free_tcp_port()
        self._test_connect([f"--bind-tcp=0.0.0.0:{port},auth=allow"], [], b"hello",
                           f"tcp://127.0.0.1:{port}/", ExitCode.OK)
        port = get_free_tcp_port()
        self._test_connect([f"--bind-tcp=0.0.0.0:{port},auth=allow"], [], b"hello",
                           f"ws://127.0.0.1:{port}/", ExitCode.OK)

    def test_ws_socket(self):
        port = get_free_tcp_port()
        self._test_connect([f"--bind-ws=0.0.0.0:{port},auth=allow"], [], b"hello",
                           f"ws://127.0.0.1:{port}/", ExitCode.OK)

    def _gen_ssl(self):
        tmpdir = tempfile.mkdtemp(suffix='ssl-xpra')
        keyfile = os.path.join(tmpdir, "key.pem")
        outfile = os.path.join(tmpdir, "out.pem")
        openssl_command = [
            "openssl", "req", "-new", "-newkey", "rsa:4096", "-days", "2", "-nodes", "-x509",
            "-subj", "/C=US/ST=Denial/L=Springfield/O=Dis/CN=localhost",
            "-keyout", keyfile, "-out", outfile,
        ]
        openssl = self.run_command(openssl_command)
        assert pollwait(openssl, 20) == 0, "openssl certificate generation failed"
        # combine the two files:
        certfile = tempfile.NamedTemporaryFile(delete=False)
        os.path.join(tmpdir, "cert.pem")
        for fname in (keyfile, outfile):
            with open(fname, 'rb') as f:
                certfile.write(f.read())
        shutil.rmtree(tmpdir)
        cert_data = load_binary_file(certfile.name)
        log("generated cert data: %s", repr_ellipsized(cert_data))
        if not cert_data:
            # cannot run openssl? (happens from rpmbuild)
            raise RuntimeError("SSL test skipped, cannot run " + " ".join(openssl_command))
        return certfile

    def verify_connect(self, uri, exit_code=ExitCode.OK, *client_args):
        cmd = ["info", uri] + list(client_args)
        env = self.get_run_env()
        env["SSL_RETRY"] = "0"
        client = self.run_xpra(cmd, env=env)
        r = pollwait(client, CONNECT_WAIT)
        if client.poll() is None:
            client.terminate()
        r = client.poll()
        self.verify_exitcode(expected=exit_code, actual=r)

    def verify_exitcode(self, client="info client", expected: ExitValue = ExitCode.OK,
                        actual: ExitValue | None = ExitCode.OK):
        if actual is None:
            raise Exception(f"expected {client} to return %s but it is still running" % (estr(expected),))
        if actual != expected:
            raise RuntimeError(f"expected {client} to return %s but got %s" % (estr(expected), estr(actual)))

    def test_quic_socket(self):
        port = get_free_tcp_port()
        try:
            from xpra.net.quic import listener, client
            assert listener and client
        except ImportError as e:
            print(f"quic socket test skipped: {e}")
            return
        display_no = self.find_free_display_no()
        display = f":{display_no}"
        with GenSSLCertContext() as genssl:
            server = None
            try:
                log("starting test quic server on %s", display)
                server = self.start_server(display,
                                           f"--bind-quic=0.0.0.0:{port},auth=allow",
                                           f"--ssl-cert={genssl.certfile}",
                                           f"--ssl-key={genssl.keyfile}",
                                           )

                def tc(exit_code, *client_args):
                    self.verify_connect(f"quic://127.0.0.1:{port}/", exit_code, *client_args)

                # asyncio makes it too difficult to emit the correct exception here:
                # we should be getting ExitCode.SSL_CERTIFICATE_VERIFY_FAILURE..
                tc(ExitCode.CONNECTION_FAILED)
                tc(ExitCode.OK, NOVERIFY, NOHOSTNAME)
            finally:
                if server:
                    server.terminate()

    def test_ssl(self):
        server = None
        display_no = self.find_free_display_no()
        display = f":{display_no}"
        tcp_port = get_free_tcp_port()
        ws_port = get_free_tcp_port()
        wss_port = get_free_tcp_port()
        ssl_port = get_free_tcp_port()
        proto_ports = {
            "tcp": tcp_port,
            "ws": ws_port,
            "wss": wss_port,
            "ssl": ssl_port,
        }
        ports_proto = dict((v, k) for k, v in proto_ports.items())
        with GenSSLCertContext() as genssl:
            try:
                server_args = [
                    "--ssl=on",
                    "--html=on",
                    f"--ssl-cert={genssl.certfile}",
                    f"--ssl-key={genssl.keyfile}",
                ]
                for bind_mode, bind_port in proto_ports.items():
                    server_args.append(f"--bind-{bind_mode}=0.0.0.0:{bind_port},auth=allow")
                log("starting test ssl server on %s", display)
                server = self.start_server(display, *server_args)

                # test it with openssl client:
                for mode, verify_port in proto_ports.items():
                    openssl_verify_command = (
                        "openssl", "s_client", "-connect",
                        "127.0.0.1:%i" % verify_port, "-CAfile", genssl.certfile,
                    )
                    devnull = os.open(os.devnull, os.O_WRONLY)
                    openssl = self.run_command(openssl_verify_command, stdin=devnull, shell=True)
                    r = pollwait(openssl, 10)
                    if r != 0:
                        raise RuntimeError(f"openssl certificate returned {r} for {mode} port {verify_port}")

                errors: list[str] = []

                def tc(mode: str, port: int):
                    uri = f"{mode}://foo:bar@127.0.0.1:{port}/"
                    stype = ports_proto.get(port, "").rjust(5)
                    try:
                        self.verify_connect(uri, ExitCode.OK, NOVERIFY, NOHOSTNAME)
                    except RuntimeError as e:
                        err = f"failed to connect to {stype} port using mode {mode} with uri {uri!r} and {NOVERIFY!r} {NOHOSTNAME!r}: {e}"
                        log.error(f"Error: {err}")
                        errors.append(err)
                    # without `NOHOSTNAME`, connection should fail with a SSL failure:
                    try:
                        self.verify_connect(uri, ExitCode.SSL_FAILURE, NOVERIFY)
                    except RuntimeError as e:
                        err = f"connect to {stype} port using uri {uri} with {NOVERIFY!r}: {e}"
                        log.error(f"Error: {err}")
                        errors.append(err)
                    # without `NOVERIFY`:
                    try:
                        self.verify_connect(uri, ExitCode.SSL_CERTIFICATE_VERIFY_FAILURE, NOHOSTNAME)
                    except RuntimeError as e:
                        err = f"connect to {stype} port using uri {uri} with {NOHOSTNAME!r}: {e}"
                        log.error(f"Error: {err}")
                        errors.append(err)
                    # without any ssl options:
                    try:
                        self.verify_connect(uri, ExitCode.SSL_CERTIFICATE_VERIFY_FAILURE)
                    except RuntimeError as e:
                        err = f"connect to {stype} port using uri {uri}: {e}"
                        log.error(f"Error: {err}")
                        errors.append(err)

                # connect to ssl socket:
                tc("ssl", ssl_port)
                # tcp socket should upgrade to ssl:
                tc("ssl", tcp_port)
                # tcp socket should upgrade to ws and ssl:
                tc("wss", tcp_port)
                # ws socket should upgrade to ssl:
                tc("wss", ws_port)
                if errors:
                    log.error(f"{len(errors)} ssl test errors with server args: {server_args}")
                    msg = "\n* ".join(errors)
                    raise RuntimeError(f"{len(errors)} errors testing ssl sockets:\n* {msg}")
            finally:
                if server:
                    server.terminate()

    def test_bind_tmpdir(self):
        # remove socket dirs from default arguments temporarily:
        saved_args = ServerSocketsTest.default_xpra_args
        tmpsocketdir1 = tempfile.mkdtemp(suffix='xpra')
        tmpsocketdir2 = tempfile.mkdtemp(suffix='xpra')
        tmpsessionsdir = tempfile.mkdtemp(suffix='xpra')
        # hide sessions dir and use a single socket dir location:
        ServerSocketsTest.default_xpra_args = list(filter(lambda x: not x.startswith("--socket-dir"), saved_args))
        server_args = (
            "--socket-dir=%s" % tmpsocketdir1,
            "--socket-dirs=%s" % tmpsocketdir2,
            "--sessions-dir=%s" % tmpsessionsdir,
            "--bind=noabstract",
        )
        log_gap()

        def t(client_args=(), prefix=DISPLAY_PREFIX, exit_code=ExitCode.OK):
            self._test_connect(server_args, client_args, None, prefix, exit_code)

        try:
            # it should not be found by default
            # since we only use hidden temporary locations
            # for both sessions-dir and socket-dir(s):
            t(exit_code=ExitCode.CONNECTION_FAILED)
            # specifying the socket-dir(s) should work:
            for d in (tmpsocketdir1, tmpsocketdir2):
                t(("--socket-dir=%s" % d,))
                t(("--socket-dirs=%s" % d,))
        finally:
            ServerSocketsTest.default_xpra_args = saved_args
            for d in (tmpsocketdir1, tmpsocketdir2, tmpsessionsdir):
                shutil.rmtree(d)


def main():
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Deque
from collections.abc import Callable
from time import monotonic
from collections import deque

from xpra.common import noop
from xpra.platform.paths import get_icon_filename
from xpra.log import Logger

log = Logger("tray")


class TrayBase:
    """
        Utility superclass for all tray implementations
    """

    def __init__(self, _client, app_id, menu, tooltip: str = "", icon_filename: str = "",
                 size_changed_cb: Callable = noop, click_cb: Callable = noop,
                 mouseover_cb: Callable = noop, exit_cb: Callable = noop):
        # we don't keep a reference to client,
        # because calling functions on the client directly should be discouraged
        self.app_id = app_id
        self.menu = menu
        self.tooltip = tooltip
        self.size_changed_cb = size_changed_cb
        self.click_cb = click_cb
        self.mouseover_cb = mouseover_cb
        self.exit_cb = exit_cb
        self.tray_widget = None
        self.default_icon_filename = icon_filename  # ie: "xpra" or "/path/to/xpra.png"
        # some implementations need this for guessing the geometry (see recalculate_geometry):
        self.geometry_guess: tuple[int, int, int, int] | None = None
        self.tray_event_locations: Deque[tuple[int, int]] = deque(maxlen=512)
        self.default_icon_extension = "png"
        self.icon_timestamp = 0.0

    def __repr__(self):
        return f"Tray({self.app_id}:{self.tooltip})"

    def cleanup(self) -> None:
        if self.tray_widget:
            self.hide()
            self.tray_widget = None

    def ready(self) -> None:
        """
        This is called when we have finished the startup sequence.
        The MacOS dock overrides this method.
        """

    def show(self) -> None:
        raise NotImplementedError

    def hide(self) -> None:
        raise NotImplementedError

    def get_orientation(self):
        return ""  # assume "HORIZONTAL"

    def get_geometry(self):
        raise NotImplementedError

    def get_size(self) -> tuple[int, int] | None:
        g = self.get_geometry()
        if g is None:
            return None
        return g[2:4]

    def set_tooltip(self, tooltip: str = "") -> None:
        self.tooltip = tooltip
        raise NotImplementedError

    def set_blinking(self, on: bool) -> None:
        raise NotImplementedError

    def set_icon_from_data(self, pixels, has_alpha: bool, w: int, h: int, rowstride: int, options=None):
        raise NotImplementedError

    def get_icon_filename(self, basename="") -> str:
        name = basename or self.default_icon_filename
        f = get_icon_filename(name, self.default_icon_extension)
        if not f:
            log.error(f"Error: cannot find icon {name!r}")
        return f

    def set_icon(self, basename="") -> None:
        filename = self.get_icon_filename(basename)
        if not filename:
            return
        log(f"set_icon({basename}) using filename={filename!r}")
        self.set_icon_from_file(filename)

    def set_icon_from_file(self, filename: str) -> None:
        log(f"set_icon_from_file({filename}) tray_widget={self.tray_widget}")
        if not self.tray_widget:
            return
        self.do_set_icon_from_file(filename)
        self.icon_timestamp = monotonic()

    def do_set_icon_from_file(self, filename: str) -> None:
        raise NotImplementedError

    def recalculate_geometry(self, x: int, y: int, width: int, height: int) -> None:
        log("recalculate_geometry%s guess=%s, tray event locations: %s",
            (x, y, width, height), self.geometry_guess, len(self.tray_event_locations))
        if x is None or y is None:
            return
        if self.geometry_guess is None:
            # better than nothing!
            self.geometry_guess = x, y, width, height
        if self.tray_event_locations and self.tray_event_locations[-1] == (x, y):
            # unchanged
            log("tray event location unchanged")
            return
        self.tray_event_locations.append((x, y))
        # sets of locations that can fit together within (size,size) distance of each other:
        xs, ys = set(), set()
        xs.add(x)
        ys.add(y)
        # walk through all of them in reverse (and stop when one does not fit):
        for tx, ty in reversed(self.tray_event_locations):
            minx = min(xs)
            miny = min(ys)
            maxx = max(xs)
            maxy = max(ys)
            if (tx < minx and tx < (maxx - width)) or (tx > maxx and tx > (minx + width)):
                break  # cannot fit...
            if (ty < miny and ty < (maxy - height)) or (ty > maxy and ty > (miny + height)):
                break  # cannot fit...
            xs.add(tx)
            ys.add(ty)
        # now add some padding if needed:
        minx = min(xs)
        miny = min(ys)
        maxx = max(xs)
        maxy = max(ys)
        padx = width - (maxx - minx)
        pady = height - (maxy - miny)
        assert padx >= 0 and pady >= 0
        minx -= padx // 2
        miny -= pady // 2
        oldgeom = self.geometry_guess
        self.geometry_guess = max(0, minx), max(0, miny), width, height
        log("recalculate_geometry() geometry guess=%s (old guess=%s)", self.geometry_guess, oldgeom)
        if self.geometry_guess != oldgeom:
            self.size_changed_cb()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any


def get_pyobjc_version():
    try:
        import objc
        return objc.__version__
    except ImportError:
        return None


def get_version_info() -> dict[str, Any]:
    d = {}
    v = get_pyobjc_version()
    if v:
        d["pyobjc.version"] = v
    return d



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.net import compression


class TestCompression(unittest.TestCase):

    def test_main(self):
        compression.init_all()
        assert compression.use("lz4")
        assert compression.get_compression_caps()
        assert compression.get_enabled_compressors()
        for x in compression.get_enabled_compressors():
            assert compression.get_compressor(x)

    def _test_wrapper_class(self, clazz):
        for data in (b"hello", memoryview(b"foo")):
            x = clazz("datatype", data)
            assert repr(x)
            assert len(x) == len(data)
        x = clazz("fake", b"foo")
        assert repr(x)
        return x

    def test_compressed_class(self):
        self._test_wrapper_class(compression.Compressed)

    def test_large_structure(self):
        self._test_wrapper_class(compression.LargeStructure)

    def test_compressible(self):
        x = self._test_wrapper_class(compression.Compressible)
        try:
            x.compress()
        except Exception:
            pass
        else:
            raise Exception("compressible.compress must be overriden")

    def test_decompress_by_name(self):
        for x in ("foo", "lz4XX", "zli"):
            try:
                compression.decompress_by_name(b"foo", x)
            except Exception:
                pass
            else:
                raise Exception(f"{x} is not a valid compression")

    def test_compressed_wrapper(self):
        r = compression.compressed_wrapper("test", b"a" * max(1, (compression.MIN_COMPRESS_SIZE + 1)))
        if not r.datatype.startswith("raw"):
            raise Exception(f"should not be able to use the wrapper without enabling a compressor, but got {r!r}")
        for x in ("lz4", "brotli", "none"):
            if not compression.use(x):
                continue
            kwargs = {x: True}
            for level in (0, 1, 5, 10):
                for data in (
                        b"0" * 1024,
                        b"0" * 16,
                        b"\0",
                        memoryview(b"hello"),
                        bytearray(b"hello"),
                        b"1" * 1024 * 1024 * 16,
                ):
                    v = compression.compressed_wrapper("test", data, level=level, **kwargs)
                    assert v
                    assert repr(v)
                    assert compression.get_compression_type(v.level)
                    #and back:
                    try:
                        d = compression.decompress_by_name(v.data, v.algorithm)
                        assert d
                        if x != "none":
                            #we can't do none,
                            #because it would be mistaken for "zlib"
                            #(for historical reasons - 'zlib' uses level=0, and 'none' has no level)
                            d = compression.decompress(v.data, v.level)
                            assert d
                    except Exception:
                        print(f"error decompressing {v} - generated with settings: {kwargs}")
                        raise

    def test_lz4(self):
        try:
            from xpra.net.lz4.lz4 import compress, decompress
            from lz4 import block
        except ImportError as e:
            print(f"lz4 test skipped: {e}")
            return
        for t in (
                b"abc", b"foobar",
                b"\0" * 1000000,
        ):
            for accel in (0, 1, 5, 9):
                N = 1
                for _ in range(N):
                    c1 = compress(t, acceleration=accel)
                for _ in range(N):
                    c2 = block.compress(t, mode="fast", acceleration=accel)
                assert c1 == c2
                d1 = decompress(c1)
                d2 = block.decompress(c2)
                assert d1 == d2 == t


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, Gdk   # pylint: disable=wrong-import-position @UnresolvedImport


def main():
	display = Gdk.Display.get_default()
	#cursor from name:
	cursor = Gdk.Cursor.new_from_name(display, "xterm")
	cursors = [("new_from_name", cursor)]
	surface, xhot, yhot = cursor.get_surface()
	#from the pixbuf:
	pixbuf = cursor.get_image()
	import binascii
	print("pixbuf non-zero pixels=%s" % binascii.hexlify(pixbuf.get_pixels()).rstrip(b"0"))
	pcursor = Gdk.Cursor.new_from_pixbuf(display, pixbuf, xhot, yhot)
	cursors.append(("new_from_pixbuf", pcursor))
	w = surface.get_width()
	h = surface.get_height()
	pixbuf = Gdk.pixbuf_get_from_surface(surface, 0, 0, w, h)
	pscursor = Gdk.Cursor.new_from_pixbuf(display, pixbuf, xhot, yhot)
	cursors.append(("new_from_pixbuf from surface", pscursor))
	scursor = Gdk.Cursor.new_from_surface(display, surface, xhot, yhot)
	cursors.append(("new_from_surface", scursor))

	w = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	w.connect("delete_event", Gtk.main_quit)
	w.set_title("Cursor Pixbuf Test")
	w.set_size_request(320, 200)
	b = Gtk.Button()
	def change_cursor(*_args):
		info, cursor = cursors.pop(0)
		print("%s : %s" % (info, cursor))
		w.get_window().set_cursor(cursor)
		cursors.append((info, cursor))
	b.connect("clicked", change_cursor)
	w.add(b)
	w.show_all()
	change_cursor()
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shlex
from typing import Iterable

from xpra.os_util import gi_import, getuid
from xpra.util.system import is_distribution_variant
from xpra.util.thread import start_thread
from xpra.util.io import get_status_output, which
from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.widget import label
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("util")


SUPPORTED_RPM_DISTROS = ("Fedora", "CentOS", "RedHat", "AlmaLinux", "RockyLinux", "OracleLinux")
SUPPORTED_DEB_DISTROS = ("Debian", "Ubuntu")
SUPPORTED_DISTROS = SUPPORTED_RPM_DISTROS + SUPPORTED_DEB_DISTROS


def get_terminal() -> str:
    # solution based on https://github.com/i3/i3/blob/next/i3-sensible-terminal
    for terminal in (
        os.environ.get("TERMINAL", ""),
        "x-terminal-emulator",
        "mate-terminal", "gnome-terminal",
        "terminator", "xfce4-terminal", "urxvt", "rxvt", "termit",
        "Eterm", "aterm", "uxterm", "xterm", "roxterm", "termite",
        "lxterminal", "terminology", "st", "qterminal", "lilyterm",
        "tilix", "terminix", "konsole", "kitty", "guake", "tilda",
        "alacritty", "hyper", "wezterm", "rio",
    ):
        if terminal:
            cmd = which(terminal)
            if cmd:
                return cmd
    return ""


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        self.terminal = get_terminal()
        self.sudo = which("sudo")
        self.package_switch: dict[str, Gtk.Switch] = {}
        self.package_system = ""
        self.initial_state: dict[str, bool] = {}
        self.current_state: dict[str, bool] = {}
        self.apply_button = None
        super().__init__(
            "Install or remove Xpra Packages",
            "package.png",
            wm_class=("xpra-configure-packages-gui", "Xpra Configure Packages GUI"),
            default_size=(640, 500),
            header_bar=(False, False),
            parent=parent,
        )

    def populate(self) -> None:
        self.clear_vbox()
        self.add_widget(label("Install or remove xpra packages", font="sans 20"))

        def fail(messages: Iterable[str]) -> None:
            self.populate_form(messages, ("Exit", self.dismiss))
        if not self.terminal:
            fail((
                "Unable to locate a terminal application to use.",
                "Either install a commonly used one (ie: `xterm`),",
                "or set the `TERMINAL` environment variable.",
                ""))
            return
        if not self.sudo and getuid() != 0:
            fail((
                "Unable to locate `sudo`.",
                ""))
            return

        if not any(is_distribution_variant(x) for x in SUPPORTED_DISTROS):
            url = "https://github.com/Xpra-org/xpra/wiki/Platforms"
            fail((
                "Sorry, this tool is not supported on your distribution.",
                "",
                f"Plaase see <a href='{url}'>support platforms</a>",
                ""))
            return
        if any(is_distribution_variant(x) for x in SUPPORTED_RPM_DISTROS):
            pkgcmd = "dnf"
            querycmd = "rpm"
            self.package_system = "dnf"
        else:
            pkgcmd = "apt"
            querycmd = "dpkg"
            self.package_system = "apt"
        pkgcmd_path = which(pkgcmd)
        if not pkgcmd_path:
            fail((
                f"Unable to locate the {pkgcmd!r} command.",
                ""))
            return
        querycmd_path = which(querycmd)
        if not querycmd_path:
            fail((
                f"Unable to locate the {querycmd!r} command.",
                ""))
            return
        lines = (
            "This tool can be used to add or remove a complete set of features.",
            "This is somewhat similar to turning features on or off,",
            "but in a more global and definitive way.",
            "This requires super-user privileges via `sudo` for regular users.",
        )
        text = "\n".join(lines)
        lbl = label(text, font="Sans 14")
        lbl.set_line_wrap(True)
        lbl.set_use_markup(True)
        self.add_widget(lbl)

        grid = Gtk.Grid()
        grid.set_margin_start(40)
        grid.set_margin_end(40)
        grid.set_row_homogeneous(True)
        grid.set_column_homogeneous(False)
        self.add_widget(grid)

        for i, (package, description) in enumerate(
                {
                    "Client": "For connecting to servers",
                    "Client GTK3": "The main xpra client",
                    "HTML5": "The html5 client",
                    "Server": "For creating sessions",
                    "X11": "X11 component for both clients and servers",
                    "Audio": "Audio support: forwarding of speaker and microphone",
                    "Codecs": "Core picture compression codecs",
                    "Codecs Extras": "video compression codecs with heavier footprint and / or licensing requirements",
                    "Codecs NVidia": "extra proprietary NVidia codecs",
                }.items()
        ):
            pkg = package.lower().replace(" ", "-")
            lbl = label(package, tooltip=description)
            lbl.set_hexpand(True)
            # this causes the Switch to be displayed wrong!
            # lbl.set_halign(Gtk.Align.START)
            grid.attach(lbl, 0, i, 1, 1)
            switch = Gtk.Switch()
            switch.set_sensitive(False)
            grid.attach(switch, 1, i, 1, 1)
            self.package_switch[pkg] = switch

        self.apply_button = self.add_buttons(
            ("Exit", self.dismiss),
            ("Apply", self.apply),
        )[1]
        self.apply_button.set_sensitive(False)

        self.show_all()
        start_thread(self.query_packages, "query-packages", daemon=True)

    def query_packages(self, connect=True) -> None:
        if self.package_system == "dnf":
            cmd = "rpm -qa"
        else:
            cmd = "dpkg --list | awk '{print $2}'"
        r, out, err = get_status_output(cmd, shell=True)
        if r:
            log.error("Error: failed to retrieve the list of packages")
            log.error(f" err={err}")
            return
        package_list = sorted(out.replace("\r", "\n").split("\n"))
        log(f"{package_list=}")

        def find_package(name: str) -> bool:
            if name in package_list:
                return True
            partial = [x for x in package_list if x.startswith(name)]
            # ie: "xpra-6.0-r36666.x86_64"
            for x in partial:
                remainder = x[len(name):]
                if remainder[0] != "-":
                    continue
                # make sure the next part is a number:
                # `xpra-codecs-extras` should not match `xpra-codecs`
                # but `xpra-codecs-6.0` should match `xpra-codecs`
                if remainder[1] in "0123456789":
                    return True
            return False

        for package, switch in self.package_switch.items():
            found = find_package(f"xpra-{package}")
            self.initial_state[package] = found
            self.current_state[package] = found
            switch.set_state(found)
            if connect:
                switch.connect("state-set", self.toggle_package, package)
            switch.set_sensitive(True)

    def sync_switches(self) -> None:
        for package, switch in self.package_switch.items():
            state = self.current_state[package]
            if switch.get_state() != state:
                switch.set_state(state)

    def toggle_package(self, widget, state, package: str) -> None:
        if not widget.get_sensitive():
            return
        enabled = bool(state)
        self.current_state[package] = enabled
        # now toggle any packages depending on this one:
        # ie: xpra-codecs-extras depends on xpra-coddecs
        if enabled:
            for k in self.current_state.keys():
                if package.startswith(k) and k != package:
                    # if `xpra-codecs-extras` was enabled, enable `xpra-codecs`:
                    self.current_state[k] = True
        else:
            for k in self.current_state.keys():
                if k.startswith(package) and k != package:
                    # if `xpra-codecs` was disabled, disable `xpra-codecs-extras`:
                    self.current_state[k] = False
        self.sync_switches()
        self.apply_button.set_sensitive(self.current_state != self.initial_state)
        log("toggle_package%s", (widget, state, package))

    def apply(self, *_args) -> None:
        remove = [
            f"xpra-{x}" for x in self.initial_state.keys()
            if self.initial_state[x] and not self.current_state[x]
        ]
        install = [
            f"xpra-{x}" for x in self.initial_state.keys()
            if not self.initial_state[x] and self.current_state[x]
        ]
        log(f"apply: {remove=}, {install=}")
        command = [self.sudo] if getuid() > 0 else []
        command += [self.terminal, "-e"]
        package_commands = []
        if remove:
            package_commands.append(f"{self.package_system} remove " + " ".join(remove))
        if install:
            package_commands.append(f"{self.package_system} install " + " ".join(install))
        # ie: shell_command = "dnf remove xpra-html5;dnf install xpra-client"
        shell_command = shlex.quote(";".join(package_commands))
        command += [f"bash -c {shell_command}"]
        log(f"{command=}")

        def update_packages() -> None:
            r = get_status_output(command)[0]
            log(f"update_packages() {command=} returned {r}")
            self.query_packages(False)
        self.apply_button.set_sensitive(False)
        start_thread(update_packages, "update-packages", daemon=True)


def main(_args) -> int:
    from xpra.gtk.configure.common import run_gui
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from xpra.client.mixins import encodings
from unit.test_util import silence_error
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class EncodingClientTest(ClientMixinTest):

	def test_encoding(self):
		opts = AdHocStruct()
		opts.encoding = ""
		opts.encodings = ["rgb", "png", "jpeg"]
		opts.quality = 1
		opts.min_quality = 20
		opts.speed = 0
		opts.min_speed = 20
		opts.video_scaling = "no"
		opts.video_decoders = []
		opts.csc_modules = []
		opts.video_encoders = []
		m = self._test_mixin_class(encodings.Encodings, opts, {
			"encodings" : ["rgb"],
			"encodings.core" : ["rgb32", "rgb24", "png"],
			"encodings.problematic" : [],
			"encoding" : ""
			})
		m.set_encoding("auto")
		def f(fn, err):
			try:
				fn()
			except Exception:
				pass
			else:
				raise Exception(err)
		def set_invalid_encoding():
			m.set_encoding("invalid")
		f(set_invalid_encoding, "should not be able to set encoding 'invalid'")
		#this will trigger a warning:
		with silence_error(encodings):
			m.set_encoding("jpeg")
		#quality:
		for q in (-1, 0, 1, 99, 100):
			m.quality = q
			m.send_quality()
		for q in (-2, 101):
			m.quality = q
			f(m.send_quality, "should not be able to send invalid quality %i" % q)
		#min-quality:
		for q in (-1, 0, 1, 99, 100):
			m.min_quality = q
			m.send_min_quality()
		for q in (-2, 101):
			m.min_quality = q
			f(m.send_min_quality, "should not be able to send invalid min-quality %i" % q)
		#speed:
		for s in (-1, 0, 1, 99, 100):
			m.speed = s
			m.send_speed()
		for s in (-2, 101):
			m.speed = s
			f(m.send_speed, "should not be able to send invalid speed %i" % s)
		#min-speed:
		for s in (-1, 0, 1, 99, 100):
			m.min_speed = s
			m.send_min_speed()
		for s in (-2, 101):
			m.min_speed = s
			f(m.send_min_speed, "should not be able to send invalid min-speed %i" % s)


def main():
	unittest.main()


if __name__ == '__main__':
	main()



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

# Platform-specific settings for Win32.
CAN_DAEMONIZE = False
REINIT_WINDOWS = True

CLIPBOARDS = ("CLIPBOARD",)
CLIPBOARD_GREEDY = True

SOURCE: Sequence[str] = ()

EXECUTABLE_EXTENSION = "exe"

# these don't make sense on win32:
DEFAULT_PULSEAUDIO_CONFIGURE_COMMANDS = ()
PRINT_COMMAND = ""
DEFAULT_SSH_COMMAND = "plink.exe -ssh -agent"

OPEN_COMMAND = ["start", "''"]

# not implemented:
SYSTEM_PROXY_SOCKET = "xpra-proxy"

SOCKET_OPTIONS = (
    # not supported on win32:
    # "SO_BROADCAST", "SO_RCVLOWAT",
    "SO_DONTROUTE", "SO_ERROR", "SO_EXCLUSIVEADDRUSE",
    "SO_KEEPALIVE", "SO_LINGER", "SO_OOBINLINE", "SO_RCVBUF",
    "SO_RCVTIMEO", "SO_REUSEADDR", "SO_REUSEPORT",
    "SO_SNDBUF", "SO_SNDTIMEO", "SO_TIMEOUT", "SO_TYPE",
)



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: boundscheck=False, wraparound=False

import socket as pysocket
from libc.string cimport memset   # pylint: disable=syntax-error


cdef extern from "unistd.h":
    int close(int fd)

cdef extern from "sys/socket.h":
    int AF_VSOCK
    int SOCK_DGRAM
    int SOCK_STREAM
    cdef struct sockaddr:
        pass
    ctypedef int socklen_t
    int socket(int socket_family, int socket_type, int protocol)
    int bind(int sockfd, const sockaddr *addr, socklen_t addrlen)
    int getsockname(int sockfd, sockaddr *addr, socklen_t *addrlen)
    int accept(int sockfd, sockaddr *addr, socklen_t *addrlen)
    int connect(int sockfd, const sockaddr *addr, socklen_t addrlen)


cdef extern from "linux/vm_sockets.h":
    unsigned int VMADDR_CID_ANY
    unsigned int VMADDR_CID_HYPERVISOR
    unsigned int VMADDR_CID_HOST
    unsigned int VMADDR_PORT_ANY


    unsigned int SO_VM_SOCKETS_BUFFER_SIZE
    unsigned int SO_VM_SOCKETS_BUFFER_MIN_SIZE
    unsigned int SO_VM_SOCKETS_BUFFER_MAX_SIZE
    unsigned int SO_VM_SOCKETS_PEER_HOST_VM_ID
    unsigned int SO_VM_SOCKETS_TRUSTED
    unsigned int SO_VM_SOCKETS_CONNECT_TIMEOUT
    unsigned int SO_VM_SOCKETS_NONBLOCK_TXRX

    ctypedef unsigned short __kernel_sa_family_t
    cdef struct sockaddr_vm:
        __kernel_sa_family_t svm_family
        unsigned short svm_reserved1
        unsigned int svm_port
        unsigned int svm_cid

CID_ANY = VMADDR_CID_ANY
CID_HYPERVISOR = VMADDR_CID_HYPERVISOR
CID_HOST = VMADDR_CID_HOST
PORT_ANY = VMADDR_PORT_ANY

CID_TYPES = {
             CID_ANY        : "ANY",
             CID_HYPERVISOR : "HYPERVISOR",
             CID_HOST       : "HOST",
             }
STR_TO_CID = {
              "ANY"         : CID_ANY,
              "HYPERVISOR"  : CID_HYPERVISOR,
              "HOST"        : CID_HOST,
              }

SOCK_TYPES = {
              SOCK_STREAM   : "STREAM",
              SOCK_DGRAM    : "DGRAM",
              }


def bind_vsocket(sock_type=SOCK_STREAM, cid=VMADDR_CID_HOST, port=VMADDR_PORT_ANY):
    from xpra.log import Logger
    log = Logger("network")
    log("server_socket(%s)", (SOCK_TYPES.get(sock_type, sock_type), CID_TYPES.get(cid, cid), port))
    assert sock_type in (SOCK_STREAM, SOCK_DGRAM), "invalid socket type %s" % sock_type
    #assert cid in (VMADDR_CID_ANY, VMADDR_CID_HYPERVISOR, VMADDR_CID_HOST), "invalid cid %s" % cid
    assert port==VMADDR_PORT_ANY or (port>0 and port<65536)
    log("socket(AF_VSOCK, %s, 0)", SOCK_TYPES.get(sock_type, sock_type))
    cdef int sockfd = socket(AF_VSOCK, sock_type, 0)
    log("socket(..)=%i", sockfd)
    if sockfd<0:
        raise RuntimeError("AF_VSOCK not supported")
    cdef sockaddr_vm vmsock
    memset(&vmsock, 0, sizeof(sockaddr_vm))
    vmsock.svm_family = AF_VSOCK
    vmsock.svm_cid = cid    #VMADDR_CID_HOST
    vmsock.svm_port = port

    if bind(sockfd, <sockaddr*> &vmsock, sizeof(sockaddr_vm)):
        close(sockfd)
        raise RuntimeError("failed to bind to AF_VSOCK socket %i:%i", cid, port)

    cdef socklen_t socklen = sizeof(sockaddr_vm)
    if getsockname(sockfd, <sockaddr *> &vmsock, &socklen):
        close(sockfd)
        raise RuntimeError("getsockname failed")
    log("cid=%s, port=%i", CID_TYPES.get(vmsock.svm_cid, vmsock.svm_cid), vmsock.svm_port)
    vsock = VSocket(sockfd)
    return vsock

class VSocket:
    def __init__(self, sockfd):
        self.sockfd = sockfd
        self.sock = pysocket.fromfd(sockfd, AF_VSOCK, 0)
        self.address = None

    def __getattr__(self, attr):
        return getattr(self.sock, attr)

    def accept(self):
        cdef sockaddr_vm vmsock
        cdef socklen_t socklen = sizeof(sockaddr_vm)
        memset(&vmsock, 0, socklen)
        cdef int fd = accept(self.sockfd, <sockaddr*> &vmsock, &socklen)
        if fd<0:
            raise RuntimeError("accept failed: %s" % fd)
        self.address = (vmsock.svm_cid, vmsock.svm_port)
        conn = pysocket.fromfd(fd, AF_VSOCK, 0)
        return VSocket(conn.fileno()), self.address

    def getsockname(self):
        return self.address

    def __repr__(self):
        return "VSocket(%s)" % self.sockfd


def connect_vsocket(sock_type=SOCK_STREAM, cid=VMADDR_CID_ANY, port=VMADDR_PORT_ANY):
    from xpra.log import Logger
    log = Logger("network")
    log("connect_vsocket(%s)", (cid, port, sock_type))
    assert sock_type in (SOCK_STREAM, SOCK_DGRAM), "invalid socket type %s" % sock_type
    #assert cid in (VMADDR_CID_ANY, VMADDR_CID_HYPERVISOR, VMADDR_CID_HOST), "invalid cid %s" % cid
    assert port==VMADDR_PORT_ANY or (port>0 and port<65536)
    log("socket(%i, %i, 0)", AF_VSOCK, sock_type)
    cdef int sockfd = socket(AF_VSOCK, sock_type, 0)
    log("socket(AF_VSOCK, SOCK_DGRAM, 0)=%i", sockfd)
    if sockfd<0:
        raise RuntimeError("AF_VSOCK not supported")

    cdef sockaddr_vm vmsock
    memset(&vmsock, 0, sizeof(sockaddr_vm))
    vmsock.svm_family = AF_VSOCK
    vmsock.svm_cid = cid
    vmsock.svm_port = port

    if connect(sockfd, <sockaddr *> &vmsock, sizeof(sockaddr_vm)):
        raise RuntimeError("failed to connect to server vsock %i:%i" % (cid, port))

    vsock = VSocket(sockfd)
    return vsock



#!/usr/bin/env python3

import csv
import collections

#----------------------------------------------------------------
# This is a specially modified chart generator that takes many
# repetions of the same test and displays them all on each chart.
#----------------------------------------------------------------

#
#----------------------------------------------------------------
# The files this generator acts upon are the CSV files output
# from one or more runs of test_measure_perf.py.
#
# Data file naming convention: prefix_id_rep.csv
#
# This script takes no arguments. When it's run, it will
# produce an HTML file called "test_perf_charts.html".
#
# Open that file in your browser to see the charts.
#----------------------------------------------------------------

#----------------------------------------------------------------
# The following variables will all need to be edited to match
# the tests that you are charting
#----------------------------------------------------------------
#
# vpx is now vp8
# x264 is now h264
#
# TODO
# -- Move the legend into the chart title area
# -- Abbreviate the app names so they fit on the chart
# -- Update the documentation to describe how to use multiple unique
#    directories containing files with same key, instead of one directory
#    and files with unique keys.
#

# Location of the data files
base_dir = "/home/nickc/xtests/logs"

# Data file prefix
prefix_all = "all_tests_40"
prefix_glx = "h264_glx"

# Result subdirectories
subs_h = ["hv", "h1", "h2", "h3"]
subs_8 = ["0.15.0", "8585_1", "8585_2", "8585_3"]
subs_9 = ["0.15.0", "9612_1", "9612_2", "9612_3"]

# id is the actual id string used in the data file name
# dir is an optional subdirectory within the base_dir where the data is stored
# display is how that parameter should be displayed in the charts
params = [
    {"id": "8585", "dir": subs_8[0], "display": "8585_all_0", "prefix": prefix_all},
    {"id": "8585", "dir": subs_8[1], "display": "8585_all_1", "prefix": prefix_all},
    {"id": "8585", "dir": subs_8[2], "display": "8585_all_2", "prefix": prefix_all},
    {"id": "8585", "dir": subs_8[3], "display": "8585_all_3", "prefix": prefix_all},
    {"id": "8585", "dir": subs_h[0], "display": "8585_glx_0", "prefix": prefix_glx},
    {"id": "8585", "dir": subs_h[1], "display": "8585_glx_1", "prefix": prefix_glx},
    {"id": "8585", "dir": subs_h[2], "display": "8585_glx_2", "prefix": prefix_glx},
    {"id": "8585", "dir": subs_h[3], "display": "8585_glx_3", "prefix": prefix_glx},
    {"id": "9612", "dir": subs_9[0], "display": "9612_all_0", "prefix": prefix_all},
    {"id": "9612", "dir": subs_9[1], "display": "9612_all_1", "prefix": prefix_all},
    {"id": "9612", "dir": subs_9[2], "display": "9612_all_2", "prefix": prefix_all},
    {"id": "9612", "dir": subs_9[3], "display": "9612_all_3", "prefix": prefix_all},
    {"id": "9612", "dir": subs_h[0], "display": "9612_glx_0", "prefix": prefix_glx},
    {"id": "9612", "dir": subs_h[1], "display": "9612_glx_1", "prefix": prefix_glx},
    {"id": "9612", "dir": subs_h[2], "display": "9612_glx_2", "prefix": prefix_glx},
    {"id": "9612", "dir": subs_h[3], "display": "9612_glx_3", "prefix": prefix_glx}
]
#params = [
#    {"id": "8585", "display": "8585"},
#    {"id": "9612", "display": "9612"}
#]

# The description will be shown on the output page
description = 'Results of every test run. The red bars are from the original 2 comparisons.'

# Each file name's 'rep' value is the sequence number of that
# data file, when results of multiple files should be averaged
reps = 9     # Number of data files in each set

# Options
needReallyUniqueId = True
showLegend = "false"

#----------------------------------------------------------------
# Set any of the values in the following lists to 1 in order to
# include that test app, or metric column in the chart page.
#
apps = {"glxgears": 0,
        "glxspheres": 1,
        "moebiusgears": 0,
        "polytopes": 0,
        "x11perf": 0,
        "xterm": 0,
        "gtkperf": 0}

metrics = {"Regions/s": 1,
           "Pixels/s Sent": 1,
           "Encoding Pixels/s": 1,
           "Decoding Pixels/s": 1,
           "Application packets in/s": 1,
           "Application bytes in/s": 1,
           "Application packets out/s": 1,
           "Application bytes out/s": 1,
           "client user cpu_pct": 1,
           "client system cpu pct": 1,
           "client number of threads": 1,
           "client vsize (MB)": 1,
           "client rss (MB)": 1,
           "server user cpu_pct": 1,
           "server system cpu pct": 1,
           "server number of threads": 1,
           "server vsize (MB)": 1,
           "server rss (MB)": 1,
           "Min Batch Delay (ms)": 1,
           "Avg Batch Delay (ms)": 1,
           "Max Batch Delay (ms)": 1,
           "Min Damage Latency (ms)": 1,
           "Avg Damage Latency (ms)": 1,
           "Max Damage Latency (ms)": 1,
           "Min Quality": 1,
           "Avg Quality": 1,
           "Max Quality": 1,
           "Min Speed": 0,
           "Avg Speed": 0,
           "Max Speed": 0}

encodings = {"png": 0,
             "rgb24": 0,
             "jpeg": 0,
             "h264": 1,
             "vp8": 0,
             "vp9": 0,
             "mmap": 0}

header_dupes = []
headers = {}
titles = []
param_ids = []
param_names = []

ENCODING_RGB24 = "rgb24"

def tree():
    return collections.defaultdict(tree)
tests = tree()

def ftree():
    return collections.defaultdict(float)

# Create test map -- schema:
# {metric: {encoding: {id: {app: {rep: avg_value}}}}}
def accumulate_values(file_name, rep, param, uniqueId):
    rownum = 0
    rgb_count = 0
    rgb_values = None

    ifile = open(file_name, "rb")
    for row in csv.reader(ifile, skipinitialspace=True):
        if (rownum == 0):
            if (len(headers) == 0):
                get_headers(row)
        else:
            app = get_value(row, "Test Command")
            if (apps[app] == 1):
                encoding = get_value(row, "Encoding")
                # x264 is now h264
                if (encoding == 'x264'):
                    encoding = 'h264'
                # vpx is now vp8
                if (encoding == 'vpx'):
                    encoding = 'vp8'

                if (encodings[encoding] == 1):
                    if (encoding == ENCODING_RGB24):
                        if (rgb_values is None):
                            rgb_values = ftree()
                            rgb_count = 0

                    for metric in metrics:
                        if (metrics[metric] == 1):
                            row_value = float(get_metric(row, metric))
                            if (encoding == ENCODING_RGB24):
                                if (metric in rgb_values.keys()):
                                    rgb_values[metric] += row_value
                                else:
                                    rgb_values[metric] = row_value
                            else:
                                tests[metric][encoding][uniqueId][app][rep] = row_value

                    if (encoding == ENCODING_RGB24):
                        rgb_count += 1
                        if (rgb_count == 3):
                            for metric in metrics:
                                if (metrics[metric] == 1):
                                    tests[metric][encoding][uniqueId][app][rep] = rgb_values[metric] / 3
                            rgb_count = 0
                            rgb_values = None
        rownum += 1
    ifile.close()

def write_html():
    app_count = 0
    for app in apps.keys():
        if (apps[app] == 1):
            app_count += 1

    chart_count = 0
    for encoding in encodings.keys():
        if (encodings[encoding] == 1):
            chart_count += 1
    box_height = 0
    if (chart_count < 4):
        box_height = 400
    elif (chart_count < 7):
        box_height = 800
    elif (chart_count < 10):
        box_height = 1200

    ofile = open("charts.html", "w")
    ofile.write('<!DOCTYPE html>\n')
    ofile.write('<html>\n')
    ofile.write('<head>\n')
    ofile.write('  <meta http-equiv="Content-Type" content="text/html; charset=utf-8">\n')
    ofile.write('  <title>Xpra Performance Results</title>\n')
    ofile.write('  <link href="css/xpra.css" rel="stylesheet" type="text/css">\n')
    ofile.write('  <script language="javascript" type="text/javascript" src="js/jquery.js"></script>\n')
    ofile.write('  <script language="javascript" type="text/javascript" src="js/jquery.flot.js"></script>\n')
    ofile.write('  <script language="javascript" type="text/javascript" src="js/jquery.flot.categories.js"></script>\n')
    ofile.write('  <script language="javascript" type="text/javascript" src="js/jquery.flot.orderbars_mod.js"></script>\n')
    ofile.write('  <script language="javascript" type="text/javascript" src="js/xpra.js"></script>\n')
    ofile.write('  <script language="javascript" type="text/javascript">\n')
    ofile.write('    var options = {canvas:true, grid: {margin: {top:50}, hoverable: true}, series: {bars: {show: true, barWidth: 0.5}}, legend: {show: ' + showLegend + '}, '
                #' xaxis: {mode: "categories", tickLength: 0, min: -0.3, max: ' + str(app_count) +'}, '
                ' xaxis: {mode: "categories", tickLength: 0, min: -6, max: 6}, '
                ' colors: ["#cc0000", "#787A40", "#9FBF8C", "#C8AB65", "#cc0000", "#787A40", "#9FBF8C", "#C8AB65", "#cc0000", "#787A40", "#9FBF8C", "#C8AB65", "#cc0000", "#787A40", "#9FBF8C", "#C8AB65"]};\n')

    m_index = 0
    m_names = []
    for metric in sorted(tests.keys()):
        m_names.append(metric)
        e_names = []
        for encoding in sorted(tests[metric].keys()):
            e_names.append(encoding)
            titles.append(metric + ' ( ' + encoding + ' )')
            for param in sorted(tests[metric][encoding].keys()):
                ofile.write('    var e' + str(m_index) + '_' + encoding + '_' + param + ' = [')
                for app in sorted(tests[metric][encoding][param].keys()):
                    value = 0
                    actual_reps = 0
                    for rep in sorted(tests[metric][encoding][param][app].keys()):
                        value += float(tests[metric][encoding][param][app][rep])
                        actual_reps += 1
                    value = value / actual_reps
                    ofile.write('["' + app + '", ' + str(value) + '], ')
                ofile.write('];' + '\n')
            varStr = '    var d' + str(m_index) + '_' + encoding + ' = ['
            for i in range(0, len(param_ids)):
                if (i > 0):
                    varStr += ','
                varStr += '{label: "' + param_names[i] + '", data: e' + str(m_index) + '_' + encoding + '_' + param_ids[i] + ', bars:{order:' + str(i) + '}}'
            varStr += '];\n'
            ofile.write(varStr)
        m_index += 1

    chart_index = 0
    m_index = 0
    ofile.write('    $(function() {\n')
    for metric in sorted(tests.keys()):
        e_index = 0
        for encoding in sorted(tests[metric].keys()):
            ofile.write('        var plot' +str(chart_index)+ ' = $.plot($("#placeholder_' + str(m_index) + '_' + str(e_index) + '"), d' + str(m_index) + '_' + e_names[e_index] + ', options);\n')
            e_index += 1
            chart_index += 1
        m_index += 1
    title_index = 0

    for metric in sorted(tests.keys()):
        for encoding in sorted(tests[metric].keys()):
            ofile.write('        set_title(' + str(title_index) + ', "' + titles[title_index] + '");\n')
            title_index += 1

    for mx in range(0, m_index):
        ofile.write('$("#metric_link_'+str(mx)+'").click(function() {$("#metric_list").scrollTop(' + str(box_height) + '*'+str(mx)+');});')
    ofile.write('    });\n')
    ofile.write('  </script>\n')
    ofile.write('  <style>.metric_box {height: ' + str(box_height) + 'px}</style>\n')
    ofile.write('</head>\n')
    ofile.write('<body>\n')
    ofile.write('  <div id="page">\n')
    ofile.write('    <div id="header_box">\n')
    ofile.write('      <div id="header">\n')
    ofile.write('        <h2>Xpra Performance Results</h2>\n')
    ofile.write('        <h3>' + description + '</h3>\n')
    ofile.write('        <div id="help_text">Click a metric to locate it in the results.</div>\n')
    ofile.write('      </div>\n')

    ofile.write('      <div id="select_box">\n')
    m_index = 0
    for metric in sorted(tests.keys()):
        ofile.write('        <div id="metric_link_' + str(m_index) + '" style="float:left;height:20px;width:200px"><a href="#">' + metric + '</a></div>\n')
        m_index += 1
    ofile.write('      </div>\n')
    ofile.write('    </div>\n')

    ofile.write('    <div style="clear:both"></div>\n')
    ofile.write('    <div id="metric_list">\n')
    m_index = 0
    for metric in sorted(tests.keys()):
        ofile.write('      <div class="metric_box" id="metric_box_' + str(m_index) + '">\n')
        ofile.write('        <div class="metric_label">' + metric + '</div>\n')
        e_index = 0
        for encoding in sorted(tests[metric].keys()):
            ofile.write('        <div class="container">\n')
            ofile.write('          <div id="placeholder_' + str(m_index) + '_' + str(e_index) + '" class="placeholder"></div>\n')
            ofile.write('        </div>\n')
            e_index += 1

        ofile.write('      </div>\n')
        m_index += 1
    ofile.write('      <div class="metric_box"></div>\n')
    ofile.write('    </div>\n')
    ofile.write('  </div>\n')
    ofile.write('</body>\n')
    ofile.write('</html>\n')
    ofile.close()

def col_index(label):
    return headers[label]

def get_value(row, label):
    return row[col_index(label)].strip()

def get_metric(row, label):
    cell = row[col_index(label)]
    if cell is None or cell is '':
        cell = '0'
    return cell.strip()

def sanitize(dirName):
    # Make the directory name valid as a javascript variable
    newName = dirName.replace('.', '_')
    return newName

def get_headers(row):
    index = 0
    for column in row:
        col = column.strip()
        if col in headers:
            header_dupes.append(col)
        headers[col] = index
        index += 1

def print_headers():
    for entry in headers:
        print(entry + " " + str(headers[entry]))
    for entry in header_dupes:
        print("Found dupe: %s" % entry)

def main():
    for param in params:
        param_id = param_name = param['id']
        if ('dir' in param.keys()):
            param_id = sanitize(param['dir'])
        if (needReallyUniqueId):
            param_id = param['id'] + '_' + sanitize(param['dir']);

        if ('display' in param.keys()):
            param_name = param['display']
        param_ids.append(param_id)
        param_names.append(param_name)

    for param in params:
        uniqueId = param['id']
        if ('dir' in param.keys()):
            uniqueId = sanitize(param['dir'])

        if (needReallyUniqueId):
            uniqueId = param['id'] + '_' + sanitize(param['dir']);

        #file_name = base_dir + '/' + param['dir'] + '/' + param['prefix'] + '_' + param['id'] + '_x.csv'
        #print file_name

        for rep in range(0, reps):
            if ('dir' in param.keys()):
                file_name = base_dir + '/' + param['dir'] + '/' + param['prefix'] + '_' + param['id'] + '_' + str(rep+1) + '.csv'
            else:
                file_name = base_dir + '/' + param['prefix'] + '_' + param['id'] + '_' + str(rep+1) + '.csv'
            accumulate_values(file_name, rep, param, uniqueId)
    write_html()
    print('\nCreated: charts.html\n')

if __name__ == "__main__":
    main()




# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
from typing import Any
from collections.abc import Callable, MutableSequence

from xpra.os_util import OSX, WIN32, gi_import
from xpra.client.gui.widget_base import ClientWidgetBase
from xpra.client.gui.window_backing_base import fire_paint_callbacks
from xpra.client.gui.window_border import WindowBorder
from xpra.net.common import PacketElement
from xpra.common import gravity_str, WORKSPACE_UNSET, WORKSPACE_NAMES, force_size_constraint
from xpra.util.parsing import scaleup_value, scaledown_value
from xpra.util.system import is_Wayland
from xpra.util.objects import typedict
from xpra.util.str_fn import std
from xpra.util.env import envint, envbool, ignorewarnings
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("window")
plog = Logger("paint")
focuslog = Logger("focus")
mouselog = Logger("mouse")
workspacelog = Logger("workspace")
keylog = Logger("keyboard")
metalog = Logger("metadata")
geomlog = Logger("geometry")
iconlog = Logger("icon")
alphalog = Logger("alpha")
scalinglog = Logger("scaling")

SIMULATE_MOUSE_DOWN = envbool("XPRA_SIMULATE_MOUSE_DOWN", True)
PROPERTIES_DEBUG = [x.strip() for x in os.environ.get("XPRA_WINDOW_PROPERTIES_DEBUG", "").split(",")]
SET_SIZE_CONSTRAINTS = envbool("XPRA_SET_SIZE_CONSTRAINTS", True)
DEFAULT_GRAVITY = envint("XPRA_DEFAULT_GRAVITY", 0)
OVERRIDE_GRAVITY = envint("XPRA_OVERRIDE_GRAVITY", 0)
FORCE_FLUSH = envbool("XPRA_FORCE_FLUSH", False)

SHOW_SPINNER_WINDOW_TYPES = set(os.environ.get("XPRA_SHOW_SPINNER_WINDOW_TYPES", "DIALOG,NORMAL,SPLASH").split(","))


def is_wm_property(name: str) -> bool:
    return any(name.startswith(prefix) for prefix in ("_MOTIF", "WM_", "_NET_WM", "_GTK_"))


class ClientWindowBase(ClientWidgetBase):

    def __init__(self, client, group_leader, watcher_pid: int, wid: int,
                 wx: int, wy: int, ww: int, wh: int, bw: int, bh: int,
                 metadata: typedict, override_redirect: bool, client_properties,
                 border: WindowBorder, max_window_size, default_cursor_data, pixel_depth: int,
                 headerbar="no"):
        log("%s%s", type(self),
            (client, group_leader, watcher_pid, wid,
             wx, wy, ww, wh, bw, bh,
             metadata, override_redirect, client_properties,
             border, max_window_size, default_cursor_data, pixel_depth,
             headerbar))
        super().__init__(client, watcher_pid, wid, metadata.boolget("has-alpha"))
        self._override_redirect = override_redirect
        self.group_leader = group_leader
        self._pos = (wx, wy)
        self._size = (ww, wh)
        self._xscale = client.xscale
        self._yscale = client.yscale
        self._client_properties = client_properties
        self._set_initial_position = metadata.boolget("set-initial-position", False)
        self._requested_position = metadata.intpair("requested-position", None)
        self.size_constraints = typedict()
        self.geometry_hints = typedict()
        self.content_type = ""
        self._fullscreen = None
        self._maximized = False
        self._above = False
        self._below = False
        self._shaded = False
        self._sticky = False
        self._skip_pager = False
        self._skip_taskbar = False
        self._iconified = False
        self._focused = False
        self._opaque_region = ()
        self.window_gravity = OVERRIDE_GRAVITY or DEFAULT_GRAVITY
        self.border = border
        self.cursor_data = None
        self.default_cursor_data = default_cursor_data
        self.max_window_size = max_window_size
        self.button_state: dict[int, bool] = {}
        self.pixel_depth = pixel_depth  # 0 for default
        # window_offset is the delta between the location of the window requested by the server,
        # and where we actually end up mapping it on the client
        # (ie: when we reposition an OR window to ensure it is visible on screen)
        self.window_offset = None
        self.pending_refresh: list[tuple[int, int, int, int]] = []
        self.headerbar = headerbar

        self.init_window(metadata)
        self.setup_window(bw, bh)
        self.update_metadata(metadata)
        self.finalize_window()

    def __repr__(self):
        return f"ClientWindow({self.wid})"

    def init_window(self, metadata: typedict) -> None:
        self._backing = None
        self._metadata = typedict()
        # used for only sending focus events *after* the window is mapped:
        self._been_mapped = False

        def wn(w):
            return WORKSPACE_NAMES.get(w, w)

        workspace = typedict(self._client_properties).intget("workspace", -1)
        workspacelog("init_window(..) workspace from client properties %s: %s", self._client_properties, wn(workspace))
        if workspace >= 0:
            # client properties override application specified workspace value on init only:
            metadata["workspace"] = workspace
        self._window_workspace = WORKSPACE_UNSET  # will get set in set_metadata if present
        self._desktop_workspace = self.get_desktop_workspace()  # pylint: disable=assignment-from-none
        workspacelog("init_window(..) workspace=%s, current workspace=%s",
                     wn(self._window_workspace), wn(self._desktop_workspace))
        if self.max_window_size and "size-constraints" not in metadata:
            # this ensures that we will set size-constraints and honour max_window_size:
            metadata.pop("workspace", None)
            metadata["size-constraints"] = {}
        # initialize gravity early:
        sc = typedict(metadata.dictget("size-constraints", {}))
        self.window_gravity = OVERRIDE_GRAVITY or sc.intget("gravity", DEFAULT_GRAVITY)
        self.set_decorated(metadata.boolget("decorations", True))

    def finalize_window(self) -> None:
        """ subclasses override this method """

    def get_info(self) -> dict[str, Any]:
        attributes = []
        for attr in (
                "fullscreen", "maximized",
                "above", "below",
                "shaded", "sticky",
                "skip-pager", "skip-taskbar",
                "iconified", "focused",
        ):
            if getattr(self, "_%s" % attr.replace("-", "_"), False):
                attributes.append(attr)
        info = super().get_info()
        info |= {
            "override-redirect": self._override_redirect,
            # "group-leader"          : self.group_leader,
            "position": self._pos,
            "size": self._size,
            "client-properties": self._client_properties,
            "set-initial-position": self._set_initial_position,
            "size-constraints": dict(self.size_constraints),
            "geometry-hints": dict(self.geometry_hints),
            "content-type": self.content_type,
            "attributes": attributes,
            "gravity": gravity_str(self.window_gravity),
            # "border"                : self.border or "",
            # cursor_data
            "max-size": self.max_window_size,
            "button-state": self.button_state,
            "offset": self.window_offset or (0, 0),
        }
        return info

    def get_desktop_workspace(self):
        return None

    def get_window_workspace(self):
        return None

    ######################################################################
    # screen scaling:
    def fsx(self, v) -> float:
        """ convert X coordinate from server to client """
        return v * self._xscale

    def fsy(self, v) -> float:
        """ convert Y coordinate from server to client """
        return v * self._yscale

    def sx(self, v) -> int:
        """ convert X coordinate from server to client """
        return round(self.fsx(v))

    def sy(self, v) -> int:
        """ convert Y coordinate from server to client """
        return round(self.fsy(v))

    def srect(self, x, y, w, h) -> tuple[int, int, int, int]:
        """ convert rectangle coordinates from server to client """
        return self.sx(x), self.sy(y), self.sx(w), self.sy(h)

    def sp(self, x, y) -> tuple[int, int]:
        """ convert X,Y coordinates from server to client """
        return self.sx(x), self.sy(y)

    def cx(self, v) -> int:
        """ convert X coordinate from client to server """
        return round(v / self._xscale)

    def cy(self, v) -> int:
        """ convert Y coordinate from client to server """
        return round(v / self._yscale)

    def crect(self, x, y, w, h) -> tuple[int, int, int, int]:
        """ convert rectangle coordinates from client to server """
        return self.cx(x), self.cy(y), self.cx(w), self.cy(h)

    def cp(self, x, y) -> tuple[int, int]:
        """ convert X,Y coordinates from client to server """
        return self.cx(x), self.cy(y)

    def new_backing(self, bw: int, bh: int):
        backing_class = self.get_backing_class()
        log("new_backing(%s, %s) backing_class=%s", bw, bh, backing_class)
        assert backing_class is not None
        w, h = self._size
        self._backing = self.make_new_backing(backing_class, w, h, bw, bh)
        self._backing.border = self.border
        self._backing.default_cursor_data = self.default_cursor_data
        self._backing.gravity = self.window_gravity
        # this is only used by cairo to request a repaint for the fps counter:
        self._backing.repaint = self.repaint
        return self._backing._backing

    def destroy(self) -> None:
        # ensure we clear reference to other windows:
        self.group_leader = None
        self._metadata = typedict()
        if self._backing:
            self._backing.close()
            self._backing = None

    def setup_window(self, bw: int, bh: int) -> None:
        self.new_backing(bw, bh)
        # tell the server about the encoding capabilities of this backing instance:
        # but don't bother if they're the same as what we sent as defaults
        # (with a bit of magic to collapse the missing namespace from encoding_defaults)
        backing_props = self._backing.get_encoding_properties()
        encoding_defaults = self._client.encoding_defaults
        for k in tuple(backing_props.keys()):
            v = backing_props[k]
            try:
                # ie: "encodings.rgb_formats" -> "rgb_formats"
                # ie: "encoding.full_csc_modes" -> "full_csc_modes"
                ek = k.split(".", 1)[1]
            except IndexError:
                ek = k
            dv = encoding_defaults.get(ek)
            if dv is not None and dv == v:
                del backing_props[k]
        self._client_properties.update(backing_props)

    def send(self, packet_type: str, *args: PacketElement) -> None:
        self._client.send(packet_type, *args)

    def reset_icon(self) -> None:
        current_icon = self._current_icon
        iconlog("reset_icon() current icon=%s", current_icon)
        if current_icon:
            self.update_icon(current_icon)

    def update_icon(self, img) -> None:
        raise NotImplementedError

    def apply_transient_for(self, wid: int) -> None:
        raise NotImplementedError

    def paint_spinner(self, context, area) -> None:
        raise NotImplementedError

    def _pointer_modifiers(self, event):
        raise NotImplementedError

    def xget_u32_property(self, target, name):
        raise NotImplementedError

    def is_OR(self) -> bool:
        return self._override_redirect

    def update_metadata(self, metadata: typedict) -> None:
        metalog("update_metadata(%s)", metadata)
        if self._client.readonly:
            metadata.update(force_size_constraint(*self._size))
        self._metadata.update(metadata)
        try:
            self.set_metadata(metadata)
        except Exception:
            metalog.warn("failed to set window metadata to '%s'", metadata, exc_info=True)

    def _get_window_title(self, metadata) -> str:
        try:
            title = self._client.title.replace("\0", "")
            if title.find("@") < 0:
                return title
            # perform metadata variable substitutions:
            # full of Python 3 unicode headaches that don't need to be
            UNKNOWN_MACHINE = "<unknown machine>"
            remote_hostname = getattr(self._client, "_remote_hostname", None)
            remote_display = getattr(self._client, "_remote_display", None)
            if remote_display:
                # ie: "1\\WinSta0\\Default" -> 1-WinSta0-Default
                remote_display = remote_display.replace("\\", "-")
            default_values = {
                "title": "<untitled window>",
                "client-machine": UNKNOWN_MACHINE,
                "windowid": str(self.wid),
                "server-machine": std(remote_hostname) or UNKNOWN_MACHINE,
                "server-display": std(remote_display) or "<unknown display>",
            }
            metalog(f"default values: {default_values}")

            def validhostname(value):
                if value in (
                        "localhost",
                        "localhost.localdomain",
                        "",
                        None,
                ):
                    return None
                return value

            def getvar(var):
                # "hostname" is magic:
                # we try harder to find a useful value to show:
                if var in ("hostname", "hostinfo"):
                    server_display = getattr(self._client, "server_display", None)
                    if var == "hostinfo" and getattr(self._client, "mmap_enabled", False) and server_display:
                        # this is a local connection for sure and we can specify the display directly:
                        return server_display
                    import socket
                    local_hostname = validhostname(socket.gethostname())
                    # try to find the hostname:
                    proto = getattr(self._client, "_protocol", None)
                    target = None
                    if proto:
                        conn = getattr(proto, "_conn", None)
                        if conn:
                            hostname = conn.info.get("host")
                            target = str(conn.target)
                            if hostname:
                                if local_hostname == hostname and server_display:
                                    return server_display
                                return hostname
                    for m in ("client-machine", "server-machine"):
                        value = validhostname(getvar(m))
                        if value and value != local_hostname:
                            return value
                    if server_display:
                        return server_display
                    if target:
                        return target
                    return UNKNOWN_MACHINE
                value = metadata.get(var) or self._metadata.get(var)
                if value is None:
                    return default_values.get(var, "<unknown %s>" % var)
                return str(value)

            def metadata_replace(match):
                atvar = match.group(0)  # ie: '@title@'
                var = atvar[1:len(atvar) - 1]  # ie: 'title'
                if not var:
                    # atvar = "@@"
                    return "@"
                return getvar(var)

            sub = r"@[\w\-]*@"
            replaced = re.sub(sub, metadata_replace, title)
            metalog("re.sub%s=%s", (sub, metadata_replace, title), replaced)
            return replaced
        except Exception as e:
            log.error("Error parsing window title:")
            log.estr(e)
            return ""

    def set_metadata(self, metadata: typedict):
        metalog("set_metadata(%s)", metadata)
        debug_props = [x for x in PROPERTIES_DEBUG if x in metadata.keys()]
        for x in debug_props:
            metalog.info("set_metadata: %s=%s", x, metadata.get(x))
        # WARNING: "class-instance" needs to go first because others may realize the window
        # (and GTK doesn't set the "class-instance" once the window is realized)
        if "class-instance" in metadata:
            self.set_class_instance(*metadata.strtupleget("class-instance", ("xpra", "Xpra"), 2, 2))
            self.reset_icon()

        if "title" in metadata:
            title = self._get_window_title(metadata)
            self.set_title(title)

        if "icon-title" in metadata:
            icon_title = metadata.strget("icon-title")
            self.set_icon_name(icon_title)
            # the DE may have reset the icon now,
            # force it to use the one we really want:
            self.reset_icon()

        if "size-constraints" in metadata:
            sc = typedict(metadata.dictget("size-constraints", {}))
            self.size_constraints = sc
            self._set_initial_position = sc.boolget("set-initial-position", self._set_initial_position)
            self.set_size_constraints(sc, self.max_window_size)

        if "set-initial-position" in metadata:
            # this should be redundant - but we keep it here for consistency
            self._set_initial_position = metadata.boolget("set-initial-position")

        if "transient-for" in metadata:
            wid = metadata.intget("transient-for", -1)
            self.apply_transient_for(wid)

        if "modal" in metadata:
            modal = metadata.boolget("modal")
            self.set_modal(modal)

        # apply window-type hint if window has not been mapped yet:
        if "window-type" in metadata and not self.get_mapped():
            window_types = metadata.strtupleget("window-type")
            self.set_window_type(window_types)

        if "role" in metadata:
            role = metadata.strget("role")
            self.set_role(role)

        if "xid" in metadata:
            xid = metadata.strget("xid")
            self.set_xid(xid)

        if "opacity" in metadata:
            opacity = metadata.intget("opacity", -1)
            if opacity < 0:
                opacity = 1
            else:
                opacity = min(1, opacity // 0xffffffff)
            ignorewarnings(self.set_opacity, opacity)

        if "has-alpha" in metadata:
            new_alpha = metadata.boolget("has-alpha")
            if new_alpha != self._has_alpha:
                # win32 without opengl can't do transparency,
                # so it triggers too many warnings
                log_fn = alphalog.debug if WIN32 else alphalog.warn
                log_fn("Warning: window %#x changed its transparency attribute", self.wid)
                log_fn(" from %s to %s, behaviour is undefined", self._has_alpha, new_alpha)
                self._has_alpha = new_alpha

        if "maximized" in metadata:
            maximized = metadata.boolget("maximized")
            if maximized != self._maximized:
                self._maximized = maximized
                if maximized:
                    self.maximize()
                else:
                    self.unmaximize()

        if "fullscreen" in metadata:
            fullscreen = metadata.boolget("fullscreen")
            if self._fullscreen is None or self._fullscreen != fullscreen:
                self._fullscreen = fullscreen
                self.set_fullscreen(fullscreen)

        if "iconic" in metadata:
            iconified = metadata.boolget("iconic")
            if self._iconified != iconified:
                self._iconified = iconified
                if iconified:
                    self.iconify()
                else:
                    self.deiconify()

        if "decorations" in metadata:
            decorated = metadata.boolget("decorations", True)
            was_decorated = self.get_decorated()
            if WIN32 and decorated != was_decorated:
                log.info("decorations flag toggled, now %s, re-initializing window", decorated)
                GLib.idle_add(self._client.reinit_window, self.wid, self)
            else:
                self.set_decorated(metadata.boolget("decorations"))
                self.apply_geometry_hints(self.geometry_hints)

        if "above" in metadata:
            above = metadata.boolget("above")
            if self._above != above:
                self._above = above
                self.set_keep_above(above)

        if "below" in metadata:
            below = metadata.boolget("below")
            if self._below != below:
                self._below = below
                self.set_keep_below(below)

        if "shaded" in metadata:
            shaded = metadata.boolget("shaded")
            if self._shaded != shaded:
                self._shaded = shaded
                self.set_shaded(shaded)

        if "sticky" in metadata:
            sticky = metadata.boolget("sticky")
            if self._sticky != sticky:
                self._sticky = sticky
                if sticky:
                    self.stick()
                else:
                    self.unstick()

        if "skip-taskbar" in metadata:
            skip_taskbar = metadata.boolget("skip-taskbar")
            if self._skip_taskbar != skip_taskbar:
                self._skip_taskbar = skip_taskbar
                self.set_skip_taskbar_hint(skip_taskbar)

        if "skip-pager" in metadata:
            skip_pager = metadata.boolget("skip-pager")
            if self._skip_pager != skip_pager:
                self._skip_pager = skip_pager
                self.set_skip_pager_hint(skip_pager)

        if "opaque-region" in metadata:
            opaque_region = metadata.tupleget("opaque-region", None)
            if self._opaque_region != opaque_region:
                self._opaque_region = opaque_region
                self.set_opaque_region(opaque_region)

        if "workspace" in metadata:
            self.set_workspace(metadata.intget("workspace"))

        if "bypass-compositor" in metadata:
            self.set_bypass_compositor(metadata.intget("bypass-compositor"))

        if "strut" in metadata:
            self.set_strut(metadata.dictget("strut", {}))

        if "fullscreen-monitors" in metadata:
            self.set_fullscreen_monitors(metadata.inttupleget("fullscreen-monitors"))

        if "shape" in metadata:
            self.set_shape(metadata.dictget("shape", {}))

        if "command" in metadata:
            self.set_command(metadata.strget("command"))

        if "x11-property" in metadata:
            attr = metadata.tupleget("x11-property")
            if len(attr) >= 4:
                name, ptype, _, value = attr[:4]
                if not is_wm_property(name):
                    self.set_x11_property(name, ptype, value)

        if "content-type" in metadata:
            self.content_type = metadata.strget("content-type")

    def set_x11_property(self, *x11_property) -> None:
        pass  # see gtk client window base

    def set_command(self, command) -> None:
        pass  # see gtk client window base

    def set_class_instance(self, wmclass_name, wmclass_class) -> None:
        pass  # see gtk client window base

    def set_shape(self, shape) -> None:
        log("set_shape(%s) not implemented by %s", shape, type(self))

    def set_bypass_compositor(self, v) -> None:
        pass  # see gtk client window base

    def set_strut(self, strut) -> None:
        pass  # see gtk client window base

    def set_fullscreen_monitors(self, fsm) -> None:
        pass  # see gtk client window base

    def set_shaded(self, shaded) -> None:
        pass  # see gtk client window base

    def reset_size_constraints(self) -> None:
        self.set_size_constraints(self.size_constraints, self.max_window_size)

    def set_size_constraints(self, size_constraints: typedict, max_window_size: tuple[int, int]):
        if not SET_SIZE_CONSTRAINTS:
            return
        geomlog("set_size_constraints(%s, %s)", size_constraints, max_window_size)
        hints = typedict()
        client = self._client
        for (a, h1, h2) in (
                ("maximum-size", "max_width", "max_height"),
                ("minimum-size", "min_width", "min_height"),
                ("base-size", "base_width", "base_height"),
                ("increment", "width_inc", "height_inc"),
        ):
            v = size_constraints.intpair(a)
            geomlog("intpair(%s)=%s", a, v)
            if not v:
                continue
            v1, v2 = v
            if a == "maximum-size" and v1 >= 16384 and v2 >= 16384 and WIN32:
                # causes problems, see #2714, #3533
                continue
            sv1 = client.sx(v1)
            sv2 = client.sy(v2)
            if a in ("base-size", "increment"):
                # rounding is not allowed for these values
                fsv1 = client.fsx(v1)
                fsv2 = client.fsy(v2)

                def closetoint(value):
                    # tolerate some rounding error:
                    # (ie: 2:3 scaling may not give an integer without a tiny bit of rounding)
                    return abs(round(value) - value) < 0.00001

                if not closetoint(fsv1) or not closetoint(fsv2):
                    # the scaled value is not close to an int,
                    # so we can't honour it:
                    geomlog("cannot honour '%s' due to scaling, scaled values are not both integers: %s, %s",
                            a, fsv1, fsv2)
                    continue
            hints[h1] = sv1
            hints[h2] = sv2
        if not OSX:
            for (a, h) in (
                    ("minimum-aspect-ratio", "min_aspect"),
                    ("maximum-aspect-ratio", "max_aspect"),
            ):
                v = size_constraints.intpair(a)
                if v:
                    v1, v2 = v
                    hints[h] = (v1 * self._xscale) / (v2 * self._yscale)
        # apply max-size override if needed:
        w, h = max_window_size
        if w > 0 and h > 0 and not self._fullscreen:
            # get the min size, if there is one:
            minw = max(1, hints.intget("min_width", 1))
            minh = max(1, hints.intget("min_height", 1))
            # the actual max size is:
            # * greater than the min-size
            # * the lowest of the max-size set by the application and the one we have
            # * ensure we honour the other hints, and round the max-size down if needed:
            # according to the GTK docs:
            # allowed window widths are base_width + width_inc * N where N is any integer
            # allowed window heights are base_height + width_inc * N where N is any integer
            maxw = hints.intget("max_width", 32768)
            maxh = hints.intget("max_height", 32768)
            maxw = max(minw, min(w, maxw))
            maxh = max(minh, min(h, maxh))
            rw = (maxw - hints.intget("base_width", 0)) % max(hints.intget("width_inc", 1), 1)
            rh = (maxh - hints.intget("base_height", 0)) % max(hints.intget("height_inc", 1), 1)
            maxw -= rw
            maxh -= rh
            # if the hints combination is invalid, it's possible that we'll end up
            # not honouring "base" + "inc", but honouring just "min" instead:
            maxw = max(minw, maxw)
            maxh = max(minh, maxh)
            scaling_info = self._xscale, self._yscale
            geomlog("modified hints for max window size %s and scaling %s: %s (rw=%s, rh=%s) -> max=%sx%s",
                    max_window_size, scaling_info, hints, rw, rh, maxw, maxh)
            # ensure we don't have duplicates with bytes / strings,
            # and that keys are always "bytes":
            # (in practice this code should never fire, just here as a reminder)
            for x in ("max_width", "max_height"):
                hints.pop(x, None)
            # bug 2214: GTK3 on win32 gets confused if we specify a large max-size,
            # and it will mess up maximizing the window
            if not WIN32 or (maxw < 32000 or maxh < 32000):
                hints["max_width"] = maxw
                hints["max_height"] = maxh
        try:
            geomlog("calling: %s(%s)", self.apply_geometry_hints, hints)
            # save them so the window hooks can use the last value used:
            self.geometry_hints = hints
            self.apply_geometry_hints(hints)
        except Exception as e:
            geomlog("set_size_constraints%s", (size_constraints, max_window_size), exc_info=True)
            geomlog.error("Error setting window hints:")
            geomlog.estr(e)
            for k, v in hints.items():
                geomlog.error(f"  {k}={v}")
            geomlog.error(" from size constraints:")
            for k, v in size_constraints.items():
                geomlog.error("  %s=%s", k, v)
        self.window_gravity = OVERRIDE_GRAVITY or size_constraints.intget("gravity", DEFAULT_GRAVITY)
        b = self._backing
        if b:
            b.gravity = self.window_gravity

    def set_window_type(self, window_types) -> None:
        pass  # see gtk client window base

    def set_workspace(self, workspace) -> None:
        pass  # see gtk client window base

    def set_fullscreen(self, fullscreen) -> None:
        pass  # see gtk client window base

    def set_xid(self, xid) -> None:
        pass  # see gtk client window base

    def toggle_debug(self, *_args) -> None:
        b = self._backing
        log.info("toggling debug on backing %s for window %i", b, self.wid)
        if not b:
            return
        if b.paint_box_line_width > 0:
            b.paint_box_line_width = 0
        else:
            b.paint_box_line_width = b.default_paint_box_line_width

    def increase_quality(self, *_args) -> None:
        if self._client.quality > 0:
            # change fixed quality:
            self._client.quality = min(100, self._client.quality + 10)
            self._client.send_quality()
            log("new quality=%s", self._client.quality)
        else:
            self._client.min_quality = min(100, self._client.min_quality + 10)
            self._client.send_min_quality()
            log("new min-quality=%s", self._client.min_quality)

    def decrease_quality(self, *_args) -> None:
        if self._client.quality > 0:
            # change fixed quality:
            self._client.quality = max(1, self._client.quality - 10)
            self._client.send_quality()
            log("new quality=%s", self._client.quality)
        else:
            self._client.min_quality = max(0, self._client.min_quality - 10)
            self._client.send_min_quality()
            log("new min-quality=%s", self._client.min_quality)

    def increase_speed(self, *_args) -> None:
        if self._client.speed > 0:
            # change fixed speed:
            self._client.speed = min(100, self._client.speed + 10)
            self._client.send_speed()
            log("new speed=%s", self._client.speed)
        else:
            self._client.min_speed = min(100, self._client.min_speed + 10)
            self._client.send_min_speed()
            log("new min-speed=%s", self._client.min_speed)

    def decrease_speed(self, *_args) -> None:
        if self._client.speed > 0:
            # change fixed speed:
            self._client.speed = max(1, self._client.speed - 10)
            self._client.send_speed()
            log("new speed=%s", self._client.speed)
        else:
            self._client.min_speed = max(0, self._client.min_speed - 10)
            self._client.send_min_speed()
            log("new min-speed=%s", self._client.min_speed)

    def scaleup(self, *_args) -> None:
        self._client.scaleup()

    def window_scaleup(self, *_args) -> None:
        scaling = max(self._xscale, self._yscale)
        options = scaleup_value(scaling)
        scalinglog("scaleup() options=%s", options)
        if options:
            self._scaleto(min(options))

    def scaledown(self, *_args) -> None:
        self._client.scaledown()

    def window_scaledown(self, *_args) -> None:
        scaling = min(self._xscale, self._yscale)
        options = scaledown_value(scaling)
        scalinglog("scaledown() options=%s", options)
        if options:
            self._scaleto(max(options))

    def scalingoff(self) -> None:
        self._client.scalingoff()

    def window_scalingoff(self) -> None:
        self._scaleto(1)

    def _scaleto(self, scaling) -> None:
        xscale = self._xscale
        yscale = self._yscale
        w, h = self._size
        self._xscale = self._yscale = scaling
        nw = round(w / xscale * self._xscale)
        nh = round(h / yscale * self._yscale)
        self.reset_size_constraints()
        self.resize(nw, nh)

    def scalereset(self, *_args) -> None:
        self._client.scalereset()

    def magic_key(self, *args) -> None:
        b = self.border
        if b:
            b.toggle()
            log("magic_key%s border=%s", args, b)
            self.repaint(0, 0, *self._size)

    def repaint(self, x: int, y: int, w: int, h: int):
        # self.queue_draw_area(0, 0, *self._size)
        raise NotImplementedError("no repaint on %s" % type(self))

    def refresh_window(self, *args) -> None:
        log("refresh_window(%s) wid=%s", args, self.wid)
        self._client.send_refresh(self.wid)

    def refresh_all_windows(self, *_args) -> None:
        # this method is only here because we may want to fire it
        # from a --key-shortcut action and the event is delivered to
        # the "ClientWindow"
        self._client.send_refresh_all()

    def draw_region(self, x: int, y: int, width: int, height: int,
                    coding: str, img_data, rowstride: int,
                    options: typedict, callbacks: MutableSequence[Callable]):
        """ Note: this runs from the draw thread (not UI thread) """
        backing = self._backing
        if not backing:
            log("draw_region: window %s has no backing, gone?", self.wid)
            fire_paint_callbacks(callbacks, -1, "no backing")
            return
        # only register this callback if we actually need it:
        if backing.draw_needs_refresh:
            if not backing.repaint_all:
                self.pending_refresh.append((x, y, width, height))
            if options.intget("flush", 0) == 0 or FORCE_FLUSH:
                callbacks.append(self.after_draw_refresh)
        if coding == "void":
            fire_paint_callbacks(callbacks)
            return
        backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)

    def after_draw_refresh(self, success, message="") -> None:
        backing = self._backing
        plog(f"after_draw_refresh({success}, {message!r}) pending_refresh={self.pending_refresh}, {backing=}")
        if not backing:
            return
        if backing.repaint_all or self._xscale != 1 or self._yscale != 1 or is_Wayland():
            # easy: just repaint the whole window:
            rw, rh = self.get_size()
            GLib.idle_add(self.repaint, 0, 0, rw, rh)
            return
        pr = self.pending_refresh
        self.pending_refresh = []
        for x, y, w, h in pr:
            rx, ry, rw, rh = self._client.srect(x, y, w, h)
            if self.window_offset:
                rx += self.window_offset[0]
                ry += self.window_offset[1]
            GLib.idle_add(self.repaint, rx, ry, rw, rh)

    def eos(self) -> None:
        """ Note: this runs from the draw thread (not UI thread) """
        backing = self._backing
        if backing:
            backing.eos()

    def spinner(self, _ok) -> None:
        if not self.can_have_spinner():
            return
        log("spinner(%s) queueing redraw")
        # with normal windows, we just queue a draw request
        # and let the expose event paint the spinner
        w, h = self.get_size()
        self.repaint(0, 0, w, h)

    def can_have_spinner(self) -> bool:
        if self._backing is None:
            return False
        window_types = self._metadata.strtupleget("window-type")
        return bool(set(window_types) & SHOW_SPINNER_WINDOW_TYPES)

    def _focus(self) -> bool:
        return self._client.update_focus(self.wid, True)

    def _unfocus(self) -> bool:
        return self._client.update_focus(self.wid, False)

    def quit(self) -> None:
        self._client.quit(0)

    def void(self) -> None:
        """
        This method can be used to capture key shortcuts
        without triggering any specific action.
        """

    def show_window_info(self, *_args) -> None:
        from xpra.client.gtk3.window_info import WindowInfo
        wi = WindowInfo(self._client, self)
        wi.show()

    def show_session_info(self, *args) -> None:
        self._client.show_session_info(*args)

    def show_menu(self, *args) -> None:
        self._client.show_menu(*args)

    def show_start_new_command(self, *args) -> None:
        self._client.show_start_new_command(*args)

    def show_bug_report(self, *args) -> None:
        self._client.show_bug_report(*args)

    def show_file_upload(self, *args) -> None:
        self._client.show_file_upload(*args)

    def show_shortcuts(self, *args) -> None:
        self._client.show_shortcuts(*args)

    def show_docs(self, *args) -> None:
        self._client.show_docs(*args)

    @staticmethod
    def log(message="") -> None:
        log.info(message)

    def next_keyboard_layout(self, update_platform_layout) -> None:
        self._client.next_keyboard_layout(update_platform_layout)

    def keyboard_layout_changed(self, *args) -> None:
        # used by win32 hooks to tell us about keyboard layout changes for this window
        keylog("keyboard_layout_changed%s", args)
        self._client.window_keyboard_layout_changed(self)

    def get_mouse_event_wid(self, *_args) -> int:
        # used to be overridden in GTKClientWindowBase
        return self.wid

    def _do_motion_notify_event(self, event) -> None:
        if self._client.readonly or self._client.server_readonly or not self._client.server_pointer:
            return
        pointer_data, modifiers, buttons = self._pointer_modifiers(event)
        wid = self.get_mouse_event_wid(*pointer_data)
        mouselog("do_motion_notify_event(%s) wid=%s / focus=%s / window wid=%i",
                 event, wid, self._client._focused, self.wid)
        mouselog(" device=%s, pointer=%s, modifiers=%s, buttons=%s",
                 self._device_info(event), pointer_data, modifiers, buttons)
        device_id = 0
        self._client.send_mouse_position(device_id, wid, pointer_data, modifiers, buttons)

    def _device_info(self, event) -> str:
        try:
            return event.device.get_name()
        except AttributeError:
            return ""

    def _button_action(self, button: int, event, depressed: bool, props=None) -> None:
        if self._client.readonly or self._client.server_readonly or not self._client.server_pointer:
            return
        pointer_data, modifiers, buttons = self._pointer_modifiers(event)
        wid = self.get_mouse_event_wid(*pointer_data)
        mouselog("_button_action(%s, %s, %s) wid=%s / focus=%s / window wid=%i",
                 button, event, depressed, wid, self._client._focused, self.wid)
        mouselog(" device=%s, pointer=%s, modifiers=%s, buttons=%s",
                 self._device_info(event), pointer_data, modifiers, buttons)
        device_id = 0

        def send_button(pressed, **kwargs):
            sprops = props or {}
            sprops.update(kwargs)
            self._client.send_button(device_id, wid, button, pressed, pointer_data, modifiers, buttons, sprops)

        pressed_state = self.button_state.get(button, False)
        if SIMULATE_MOUSE_DOWN and pressed_state is False and depressed is False:
            mouselog("button action: simulating missing mouse-down event for window %s before mouse-up", wid)
            # (needed for some dialogs on win32):
            send_button(True, synthetic=True)
        self.button_state[button] = depressed
        send_button(depressed)

    def do_button_press_event(self, event) -> None:
        self._button_action(event.button, event, True)

    def do_button_release_event(self, event) -> None:
        self._button_action(event.button, event, False)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger

log = Logger("dbus")


def dbus_exception_wrap(fn, info="cannot setup dbus instance"):
    try:
        v = fn()
        log(f"dbus_exception_wrap() {fn}()={v}")
        return v
    except ImportError as e:
        log("%s", exc_info=True)
        log.warn(f"Warning: {info}")
        log.warn(f" {e}")
    except Exception as e:
        log("%s", exc_info=True)
        if str(e).find("org.freedesktop.DBus.Error.NoServer") < 0:
            log.error("dbus server error", exc_info=True)
        log.error(f"Error: {info}")
        # split on ":" unless it is quoted:
        tmp = "-XX-" * 10
        msg = str(e).replace("':'", tmp)
        for x in msg.split(":"):
            log.error("  %s", x.replace(tmp, "':'").strip())
    return None
